@@ -1,7 +1,7 @@
 import numbers
 import warnings
 
-import toolz
+import tlz as toolz
 
 from .. import base, utils
 from ..delayed import unpack_collections
@@ -23,7 +23,7 @@ def blockwise(
     meta=None,
     **kwargs
 ):
-    """ Tensor operation: Generalized inner and outer products
+    """Tensor operation: Generalized inner and outer products
 
     A broad class of blocked algorithms and patterns can be specified with a
     concise multi-index notation.  The ``blockwise`` function applies an in-memory
@@ -139,7 +139,7 @@ def blockwise(
     if new:
         raise ValueError("Unknown dimension", new)
 
-    from .core import Array, unify_chunks, normalize_arg
+    from .core import unify_chunks, normalize_arg
 
     if align_arrays:
         chunkss, arrays = unify_chunks(*args)
@@ -160,30 +160,35 @@ def blockwise(
         if not isinstance(v, tuple):
             v = (v,)
         chunkss[k] = v
-    arginds = list(zip(arrays, args[1::2]))
 
-    for arg, ind in arginds:
-        if hasattr(arg, "ndim") and hasattr(ind, "__len__") and arg.ndim != len(ind):
-            raise ValueError(
-                "Index string %s does not match array dimension %d" % (ind, arg.ndim)
-            )
-
-    numblocks = {a.name: a.numblocks for a, ind in arginds if ind is not None}
+    arginds = zip(arrays, args[1::2])
+    numblocks = {}
 
     dependencies = []
     arrays = []
 
     # Normalize arguments
     argindsstr = []
-    for a, ind in arginds:
+
+    for arg, ind in arginds:
         if ind is None:
-            a = normalize_arg(a)
-            a, collections = unpack_collections(a)
+            arg = normalize_arg(arg)
+            arg, collections = unpack_collections(arg)
             dependencies.extend(collections)
         else:
-            arrays.append(a)
-            a = a.name
-        argindsstr.extend((a, ind))
+            if (
+                hasattr(arg, "ndim")
+                and hasattr(ind, "__len__")
+                and arg.ndim != len(ind)
+            ):
+                raise ValueError(
+                    "Index string %s does not match array dimension %d"
+                    % (ind, arg.ndim)
+                )
+            numblocks[arg.name] = arg.numblocks
+            arrays.append(arg)
+            arg = arg.name
+        argindsstr.extend((arg, ind))
 
     # Normalize keyword arguments
     kwargs2 = {}
@@ -243,12 +248,12 @@ def blockwise(
         from .utils import compute_meta
 
         meta = compute_meta(func, dtype, *args[::2], **kwargs)
-    if meta is not None:
-        return Array(graph, out, chunks, meta=meta)
-    else:
-        return Array(graph, out, chunks, dtype=dtype)
+    return new_da_object(graph, out, chunks, meta=meta, dtype=dtype)
 
 
 def atop(*args, **kwargs):
     warnings.warn("The da.atop function has moved to da.blockwise")
     return blockwise(*args, **kwargs)
+
+
+from .core import new_da_object
