@@ -7,9 +7,11 @@ import functools
 import bisect
 
 import numpy as np
-from toolz import memoize, merge, pluck, concat, accumulate
+from tlz import memoize, merge, pluck, concat, accumulate
 
 from .. import core
+from .. import config
+from .. import utils
 from ..highlevelgraph import HighLevelGraph
 from ..base import tokenize, is_dask_collection
 
@@ -31,7 +33,7 @@ def _sanitize_index_element(ind):
 
 
 def sanitize_index(ind):
-    """ Sanitize the elements for indexing along one axis
+    """Sanitize the elements for indexing along one axis
 
     >>> sanitize_index([2, 3, 5])
     array([2, 3, 5])
@@ -83,7 +85,7 @@ def sanitize_index(ind):
         raise TypeError("Invalid index type", type(ind), ind)
 
 
-def slice_array(out_name, in_name, blockdims, index):
+def slice_array(out_name, in_name, blockdims, index, itemsize):
     """
     Master function for array slicing
 
@@ -104,6 +106,8 @@ def slice_array(out_name, in_name, blockdims, index):
       This is the dask variable output name
     blockshape - iterable of integers
     index - iterable of integers, slices, lists, or None
+    itemsize : int
+        The number of bytes required for each element of the array.
 
     Returns
     -------
@@ -135,9 +139,9 @@ def slice_array(out_name, in_name, blockdims, index):
     This function works by successively unwrapping cases and passing down
     through a sequence of functions.
 
-    slice_with_newaxis - handle None/newaxis case
-    slice_wrap_lists - handle fancy indexing with lists
-    slice_slices_and_integers - handle everything else
+    slice_with_newaxis : handle None/newaxis case
+    slice_wrap_lists : handle fancy indexing with lists
+    slice_slices_and_integers : handle everything else
     """
     blockdims = tuple(map(tuple, blockdims))
 
@@ -155,13 +159,13 @@ def slice_array(out_name, in_name, blockdims, index):
     index += (slice(None, None, None),) * missing
 
     # Pass down to next function
-    dsk_out, bd_out = slice_with_newaxes(out_name, in_name, blockdims, index)
+    dsk_out, bd_out = slice_with_newaxes(out_name, in_name, blockdims, index, itemsize)
 
     bd_out = tuple(map(tuple, bd_out))
     return dsk_out, bd_out
 
 
-def slice_with_newaxes(out_name, in_name, blockdims, index):
+def slice_with_newaxes(out_name, in_name, blockdims, index, itemsize):
     """
     Handle indexing with Nones
 
@@ -177,7 +181,7 @@ def slice_with_newaxes(out_name, in_name, blockdims, index):
             where_none[i] -= n
 
     # Pass down and do work
-    dsk, blockdims2 = slice_wrap_lists(out_name, in_name, blockdims, index2)
+    dsk, blockdims2 = slice_wrap_lists(out_name, in_name, blockdims, index2, itemsize)
 
     if where_none:
         expand = expander(where_none)
@@ -202,7 +206,7 @@ def slice_with_newaxes(out_name, in_name, blockdims, index):
         return dsk, blockdims2
 
 
-def slice_wrap_lists(out_name, in_name, blockdims, index):
+def slice_wrap_lists(out_name, in_name, blockdims, index, itemsize):
     """
     Fancy indexing along blocked array dasks
 
@@ -211,8 +215,8 @@ def slice_wrap_lists(out_name, in_name, blockdims, index):
     See Also
     --------
 
-    take - handle slicing with lists ("fancy" indexing)
-    slice_slices_and_integers - handle slicing with slices and integers
+    take : handle slicing with lists ("fancy" indexing)
+    slice_slices_and_integers : handle slicing with slices and integers
     """
     assert all(isinstance(i, (slice, list, Integral, np.ndarray)) for i in index)
     if not len(blockdims) == len(index):
@@ -244,7 +248,7 @@ def slice_wrap_lists(out_name, in_name, blockdims, index):
     if all(isinstance(i, np.ndarray) or i == slice(None, None, None) for i in index):
         axis = where_list[0]
         blockdims2, dsk3 = take(
-            out_name, in_name, blockdims, index[where_list[0]], axis=axis
+            out_name, in_name, blockdims, index[where_list[0]], itemsize, axis=axis
         )
     # Mixed case. Both slices/integers and lists. slice/integer then take
     else:
@@ -261,7 +265,7 @@ def slice_wrap_lists(out_name, in_name, blockdims, index):
         )
 
         # Do work
-        blockdims2, dsk2 = take(out_name, tmp, blockdims2, index[axis], axis=axis2)
+        blockdims2, dsk2 = take(out_name, tmp, blockdims2, index[axis], 8, axis=axis2)
         dsk3 = merge(dsk, dsk2)
 
     return dsk3, blockdims2
@@ -509,7 +513,7 @@ def partition_by_size(sizes, seq):
 
 
 def issorted(seq):
-    """ Is sequence sorted?
+    """Is sequence sorted?
 
     >>> issorted([1, 2, 3])
     True
@@ -522,7 +526,7 @@ def issorted(seq):
 
 
 def slicing_plan(chunks, index):
-    """ Construct a plan to slice chunks with the given index
+    """Construct a plan to slice chunks with the given index
 
     Parameters
     ----------
@@ -554,14 +558,14 @@ def slicing_plan(chunks, index):
     return out
 
 
-def take(outname, inname, chunks, index, axis=0):
-    """ Index array with an iterable of index
+def take(outname, inname, chunks, index, itemsize, axis=0):
+    """Index array with an iterable of index
 
     Handles a single index by a single list
 
     Mimics ``np.take``
 
-    >>> chunks, dsk = take('y', 'x', [(20, 20, 20, 20)], [5, 1, 47, 3], axis=0)
+    >>> chunks, dsk = take('y', 'x', [(20, 20, 20, 20)], [5, 1, 47, 3], 8, axis=0)
     >>> chunks
     ((2, 1, 1),)
     >>> dsk  # doctest: +SKIP
@@ -572,17 +576,29 @@ def take(outname, inname, chunks, index, axis=0):
 
     When list is sorted we retain original block structure
 
-    >>> chunks, dsk = take('y', 'x', [(20, 20, 20, 20)], [1, 3, 5, 47], axis=0)
+    >>> chunks, dsk = take('y', 'x', [(20, 20, 20, 20)], [1, 3, 5, 47], 8, axis=0)
     >>> chunks
     ((3, 1),)
     >>> dsk  # doctest: +SKIP
     {('y', 0): (getitem, ('x', 0), ([1, 3, 5],)),
      ('y', 2): (getitem, ('x', 2), ([7],))}
+
+    When any indexed blocks would otherwise grow larger than
+    dask.config.array.chunk-size, we might split them,
+    depending on the value of ``dask.config.slicing.split-large-chunks``.
+
+    >>> import dask
+    >>> with dask.config.set({"array.slicing.split-large-chunks": True}):
+    ...      chunks, dsk = take('y', 'x', [(1, 1, 1), (1000, 1000), (1000, 1000)],
+    ...                        [0] + [1] * 6 + [2], axis=0, itemsize=8)
+    >>> chunks
+    ((1, 3, 3, 1), (1000, 1000), (1000, 1000))
     """
+    from .core import PerformanceWarning
+
     plan = slicing_plan(chunks[axis], index)
     if len(plan) >= len(chunks[axis]) * 10:
         factor = math.ceil(len(plan) / len(chunks[axis]))
-        from .core import PerformanceWarning
 
         warnings.warn(
             "Slicing with an out-of-order index is generating %d "
@@ -590,9 +606,52 @@ def take(outname, inname, chunks, index, axis=0):
             PerformanceWarning,
             stacklevel=6,
         )
+    index = np.asarray(index)
 
-    index_lists = [idx for _, idx in plan]
-    where_index = [i for i, _ in plan]
+    # Check for chunks from the plan that would violate the user's
+    # configured chunk size.
+    nbytes = utils.parse_bytes(config.get("array.chunk-size"))
+    other_chunks = [chunks[i] for i in range(len(chunks)) if i != axis]
+    other_numel = np.prod([sum(x) for x in other_chunks])
+
+    if math.isnan(other_numel):
+        warnsize = maxsize = math.inf
+    else:
+        maxsize = math.ceil(nbytes / (other_numel * itemsize))
+        warnsize = maxsize * 5
+
+    split = config.get("array.slicing.split-large-chunks", None)
+
+    # Warn only when the default is not specified.
+    warned = split is not None
+
+    for _, index_list in plan:
+        if not warned and len(index_list) > warnsize:
+            msg = (
+                "Slicing is producing a large chunk. To accept the large\n"
+                "chunk and silence this warning, set the option\n"
+                "    >>> with dask.config.set(**{'array.slicing.split_large_chunks': False}):\n"
+                "    ...     array[indexer]\n\n"
+                "To avoid creating the large chunks, set the option\n"
+                "    >>> with dask.config.set(**{'array.slicing.split_large_chunks': True}):\n"
+                "    ...     array[indexer]"
+            )
+            warnings.warn(msg, PerformanceWarning, stacklevel=6)
+            warned = True
+
+    where_index = []
+    index_lists = []
+    for where_idx, index_list in plan:
+        index_length = len(index_list)
+        if split and index_length > maxsize:
+            index_sublist = np.array_split(
+                index_list, math.ceil(index_length / maxsize)
+            )
+            index_lists.extend(index_sublist)
+            where_index.extend([where_idx] * len(index_sublist))
+        else:
+            index_lists.append(np.array(index_list))
+            where_index.append(where_idx)
 
     dims = [range(len(bd)) for bd in chunks]
 
@@ -615,7 +674,7 @@ def take(outname, inname, chunks, index, axis=0):
 
 
 def posify_index(shape, ind):
-    """ Flip negative indices around to positive ones
+    """Flip negative indices around to positive ones
 
     >>> posify_index(10, 3)
     3
@@ -707,7 +766,7 @@ def new_blockdim(dim_shape, lengths, index):
 
 
 def replace_ellipsis(n, index):
-    """ Replace ... with slices, :, : ,:
+    """Replace ... with slices, :, : ,:
 
     >>> replace_ellipsis(4, (3, Ellipsis, 2))
     (3, slice(None, None, None), slice(None, None, None), 2)
@@ -728,7 +787,7 @@ def replace_ellipsis(n, index):
 
 
 def normalize_slice(idx, dim):
-    """ Normalize slices to canonical form
+    """Normalize slices to canonical form
 
     Parameters
     ----------
@@ -764,14 +823,15 @@ def normalize_slice(idx, dim):
 
 
 def normalize_index(idx, shape):
-    """ Normalize slicing indexes
+    """Normalize slicing indexes
 
     1.  Replaces ellipses with many full slices
     2.  Adds full slices to end of index
     3.  Checks bounding conditions
-    4.  Replaces numpy arrays with lists
-    5.  Posify's integers and lists
-    6.  Normalizes slices to canonical form
+    4.  Replace multidimensional numpy arrays with dask arrays
+    5.  Replaces numpy arrays with lists
+    6.  Posify's integers and lists
+    7.  Normalizes slices to canonical form
 
     Examples
     --------
@@ -785,9 +845,20 @@ def normalize_index(idx, shape):
     (slice(7, None, None),)
     >>> normalize_index((Ellipsis, None), (10,))
     (slice(None, None, None), None)
+    >>> normalize_index(np.array([[True, False], [False, True], [True, True]]), (3, 2))
+    (dask.array<array, shape=(3, 2), dtype=bool, chunksize=(3, 2), chunktype=numpy.ndarray>,)
     """
+    from .core import from_array
+
     if not isinstance(idx, tuple):
         idx = (idx,)
+
+    # if a > 1D numpy.array is provided, cast it to a dask array
+    if len(idx) > 0 and len(shape) > 1:
+        i = idx[0]
+        if isinstance(i, np.ndarray) and i.shape == shape:
+            idx = (from_array(i), *idx[1:])
+
     idx = replace_ellipsis(len(shape), idx)
     n_sliced_dims = 0
     for i in idx:
@@ -820,7 +891,7 @@ def normalize_index(idx, shape):
 
 
 def check_index(ind, dimension):
-    """ Check validity of index for a given dimension
+    """Check validity of index for a given dimension
 
     Examples
     --------
@@ -890,7 +961,7 @@ def check_index(ind, dimension):
 
 
 def slice_with_int_dask_array(x, index):
-    """ Slice x with at most one 1D dask arrays of ints.
+    """Slice x with at most one 1D dask arrays of ints.
 
     This is a helper function of :meth:`Array.__getitem__`.
 
@@ -917,7 +988,7 @@ def slice_with_int_dask_array(x, index):
         for idx in index
     ]
     if sum(fancy_indexes) > 1:
-        raise NotImplementedError("Don't yet support nd fancy indexing)")
+        raise NotImplementedError("Don't yet support nd fancy indexing")
 
     out_index = []
     dropped_axis_cnt = 0
@@ -943,7 +1014,7 @@ def slice_with_int_dask_array(x, index):
 
 
 def slice_with_int_dask_array_on_axis(x, idx, axis):
-    """ Slice a ND dask array with a 1D dask arrays of ints along the given
+    """Slice a ND dask array with a 1D dask arrays of ints along the given
     axis.
 
     This is a helper function of :func:`slice_with_int_dask_array`.
@@ -1007,7 +1078,7 @@ def slice_with_int_dask_array_on_axis(x, idx, axis):
 
 
 def slice_with_bool_dask_array(x, index):
-    """ Slice x with one or more dask arrays of bools
+    """Slice x with one or more dask arrays of bools
 
     This is a helper function of `Array.__getitem__`.
 
