@@ -225,12 +225,15 @@ def test_filled():
 
 def assert_eq_ma(a, b):
     res = a.compute()
-    assert type(res) == type(b)
-    if hasattr(res, "mask"):
-        np.testing.assert_equal(res.mask, b.mask)
-        a = da.ma.filled(a)
-        b = np.ma.filled(b)
-    assert_eq(a, b, equal_nan=True)
+    if res is np.ma.masked:
+        assert res is b
+    else:
+        assert type(res) == type(b)
+        if hasattr(res, "mask"):
+            np.testing.assert_equal(res.mask, b.mask)
+            a = da.ma.filled(a)
+            b = np.ma.filled(b)
+        assert_eq(a, b, equal_nan=True)
 
 
 @pytest.mark.parametrize("dtype", ("i8", "f8"))
@@ -261,6 +264,20 @@ def test_reductions(dtype, reduction):
     )
 
 
+@pytest.mark.parametrize("dtype", ("i8", "f8"))
+@pytest.mark.parametrize(
+    "reduction", ["sum", "prod", "mean", "var", "std", "min", "max", "any", "all"]
+)
+def test_reductions_allmasked(dtype, reduction):
+    x = np.ma.masked_array([1, 2], mask=True)
+    dx = da.from_array(x, asarray=False)
+
+    dfunc = getattr(da, reduction)
+    func = getattr(np, reduction)
+
+    assert_eq_ma(dfunc(dx), func(x))
+
+
 @pytest.mark.parametrize("reduction", ["argmin", "argmax"])
 def test_arg_reductions(reduction):
     x = np.random.random((10, 10, 10))
