@@ -2,14 +2,12 @@ import itertools
 from numbers import Number
 
 import pytest
-from distutils.version import LooseVersion
 
 np = pytest.importorskip("numpy")
 
 import dask.array as da
 from dask.utils import ignoring
 from dask.array.utils import assert_eq, same_keys, AxisError, IS_NEP18_ACTIVE
-from dask.array.numpy_compat import _numpy_115
 
 
 def test_array():
@@ -224,43 +222,51 @@ def test_flip(funcname, kwargs, shape):
 
 
 @pytest.mark.parametrize(
-    "x_shape, y_shape",
+    "x_shape, y_shape, x_chunks, y_chunks",
     [
-        [(), ()],
-        [(), (7,)],
-        [(), (7, 11)],
-        [(), (7, 11, 15)],
-        [(), (7, 11, 15, 19)],
-        [(7,), ()],
-        [(7,), (7,)],
-        [(11,), (11, 7)],
-        [(15,), (7, 15, 11)],
-        [(19,), (7, 11, 19, 15)],
-        [(7, 11), ()],
-        [(7, 11), (11,)],
-        [(7, 11), (11, 7)],
-        [(11, 15), (7, 15, 11)],
-        [(15, 19), (7, 11, 19, 15)],
-        [(7, 11, 15), ()],
-        [(7, 11, 15), (15,)],
-        [(7, 11, 15), (15, 7)],
-        [(7, 11, 15), (7, 15, 11)],
-        [(11, 15, 19), (7, 11, 19, 15)],
-        [(7, 11, 15, 19), ()],
-        [(7, 11, 15, 19), (19,)],
-        [(7, 11, 15, 19), (19, 7)],
-        [(7, 11, 15, 19), (11, 19, 13)],
-        [(7, 11, 15, 19), (7, 11, 19, 15)],
+        [(), (), (), ()],
+        [(), (7,), (), ()],
+        [(), (7, 11), (), ()],
+        [(), (7, 11, 15), (), ()],
+        [(), (7, 11, 15, 19), (), ()],
+        [(7,), (), (), ()],
+        [(7,), (7,), (), ()],
+        [(11,), (11, 7), (), ()],
+        [(15,), (7, 15, 11), (), ()],
+        [(19,), (7, 11, 19, 15), (), ()],
+        [(7, 11), (), (), ()],
+        [(7, 11), (11,), (), ()],
+        [(7, 11), (11, 7), (), ()],
+        [(11, 15), (7, 15, 11), (), ()],
+        [(15, 19), (7, 11, 19, 15), (), ()],
+        [(7, 11, 15), (), (), ()],
+        [(7, 11, 15), (15,), (), ()],
+        [(7, 11, 15), (15, 7), (), ()],
+        [(7, 11, 15), (7, 15, 11), (), ()],
+        [(11, 15, 19), (7, 11, 19, 15), (), ()],
+        [(7, 11, 15, 19), (), (), ()],
+        [(7, 11, 15, 19), (19,), (), ()],
+        [(7, 11, 15, 19), (19, 7), (), ()],
+        [(7, 11, 15, 19), (11, 19, 13), (), ()],
+        [(7, 11, 15, 19), (7, 11, 19, 15), (), ()],
+        # These tests use explicitly special/disparate chunk sizes:
+        [(), (7,), (), (5,)],
+        [(), (7, 11, 15, 19), (), (1, 3, 5, 19)],
+        [(7, 11), (11, 7), (1, 1), (1, 1)],
+        [(7, 11), (11, 7), (3, 5), (4, 2)],
+        [(7, 11), (11, 7), (7, 11), (11, 7)],
+        [(11, 15, 19), (7, 11, 19, 15), (7, 7, 7), (3, 9, 9, 9)],
+        [(3, 3, 20, 30), (3, 3, 30, 20), (1, 3, 2, 6), (1, 3, 5, 10)],
     ],
 )
-def test_matmul(x_shape, y_shape):
+def test_matmul(x_shape, y_shape, x_chunks, y_chunks):
     np.random.seed(3732)
 
     x = np.random.random(x_shape)[()]
     y = np.random.random(y_shape)[()]
 
-    a = da.from_array(x, chunks=tuple((i // 2) for i in x.shape))
-    b = da.from_array(y, chunks=tuple((i // 2) for i in y.shape))
+    a = da.from_array(x, chunks=x_chunks or tuple((i // 2) for i in x.shape))
+    b = da.from_array(y, chunks=y_chunks or tuple((i // 2) for i in y.shape))
 
     expected = None
     try:
@@ -304,6 +310,31 @@ def test_tensordot_2(axes):
     assert_eq(da.tensordot(y, y, axes=axes), np.tensordot(x, x, axes=axes))
 
 
+@pytest.mark.parametrize("chunks", ["auto", (4, 6), (2, 3), (4, 3), (2, 6)])
+def test_tensordot_double_contraction_neq2(chunks):
+    # Regression test for https://github.com/dask/dask/issues/5472
+    x = np.arange(24).reshape(4, 6)
+    y = da.from_array(x, chunks=chunks)
+    assert_eq(da.tensordot(y, y, axes=2), np.tensordot(x, x, axes=2))
+
+
+def test_tensordot_double_contraction_ngt2():
+    # Regression test for https://github.com/dask/dask/issues/5472
+    x = np.arange(60.0).reshape(3, 4, 5)
+    y = np.arange(60.0).reshape(4, 5, 3)
+    u = da.from_array(x)
+    v = da.from_array(y)
+
+    assert_eq(da.tensordot(u, v, axes=2), np.tensordot(x, y, axes=2))
+
+    x = np.arange(60.0).reshape(3, 4, 5)
+    y = np.arange(60.0).reshape(4, 5, 3)
+    u = da.from_array(x, chunks=3)
+    v = da.from_array(y)
+
+    assert_eq(da.tensordot(u, v, axes=2), np.tensordot(x, y, axes=2))
+
+
 def test_tensordot_more_than_26_dims():
     ndim = 27
     x = np.broadcast_to(1, [2] * ndim)
@@ -322,7 +353,7 @@ def test_dot_method():
 
 @pytest.mark.parametrize("shape, chunks", [((20,), (6,)), ((4, 5), (2, 3))])
 def test_vdot(shape, chunks):
-    np.random.random(1337)
+    np.random.seed(1337)
 
     x = 2 * np.random.random((2,) + shape) - 1
     x = x[0] + 1j * x[1]
@@ -340,7 +371,7 @@ def test_vdot(shape, chunks):
 
 @pytest.mark.parametrize("shape1, shape2", [((20,), (6,)), ((4, 5), (2, 3))])
 def test_outer(shape1, shape2):
-    np.random.random(1337)
+    np.random.seed(1337)
 
     x = 2 * np.random.random(shape1) - 1
     y = 2 * np.random.random(shape2) - 1
@@ -624,15 +655,106 @@ def test_histogram_normed_deprecation():
 
 
 @pytest.mark.parametrize(
-    "bins, hist_range", [(None, None), (10, None), (None, (1, 10))]
+    "bins, hist_range",
+    [
+        (None, None),
+        (10, None),
+        (10, 1),
+        (None, (1, 10)),
+        (10, [0, 1, 2]),
+        (10, [0]),
+        (10, np.array([[0, 1]])),
+        (10, da.array([[0, 1]])),
+        ([[0, 1, 2]], None),
+        (np.array([[0, 1, 2]]), None),
+        (da.array([[0, 1, 2]]), None),
+    ],
 )
 def test_histogram_bin_range_raises(bins, hist_range):
     data = da.random.random(10, chunks=2)
-    with pytest.raises(ValueError) as info:
+    with pytest.raises((ValueError, TypeError)) as info:
         da.histogram(data, bins=bins, range=hist_range)
     err_msg = str(info.value)
-    assert "bins" in err_msg
-    assert "range" in err_msg
+    assert "bins" in err_msg or "range" in err_msg
+
+
+@pytest.mark.parametrize("density", [True, False])
+@pytest.mark.parametrize("weighted", [True, False])
+@pytest.mark.parametrize("non_delayed_i", [None, 0, 1])
+@pytest.mark.parametrize("delay_n_bins", [False, True])
+def test_histogram_delayed_range(density, weighted, non_delayed_i, delay_n_bins):
+    n = 100
+    v = np.random.random(n)
+    vd = da.from_array(v, chunks=10)
+
+    if weighted:
+        weights = np.random.random(n)
+        weights_d = da.from_array(weights, chunks=vd.chunks)
+
+    d_range = [vd.min(), vd.max()]
+    if non_delayed_i is not None:
+        d_range[non_delayed_i] = d_range[non_delayed_i].compute()
+    hist_d, bins_d = da.histogram(
+        vd,
+        bins=da.array(n) if delay_n_bins and not density else n,
+        range=d_range,
+        density=density,
+        weights=weights_d if weighted else None,
+    )
+
+    hist, bins = np.histogram(
+        v,
+        bins=n,
+        range=[v.min(), v.max()],
+        density=density,
+        weights=weights if weighted else None,
+    )
+
+    assert_eq(hist_d, hist)
+    assert_eq(bins_d, bins)
+
+
+@pytest.mark.parametrize("density", [True, False])
+@pytest.mark.parametrize("weighted", [True, False])
+def test_histogram_delayed_bins(density, weighted):
+    n = 100
+    v = np.random.random(n)
+    bins = np.array([0, 0.2, 0.5, 0.8, 1])
+
+    vd = da.from_array(v, chunks=10)
+    bins_d = da.from_array(bins, chunks=2)
+
+    if weighted:
+        weights = np.random.random(n)
+        weights_d = da.from_array(weights, chunks=vd.chunks)
+
+    hist_d, bins_d2 = da.histogram(
+        vd,
+        bins=bins_d,
+        range=[bins_d[0], bins_d[-1]],
+        density=density,
+        weights=weights_d if weighted else None,
+    )
+
+    hist, bins = np.histogram(
+        v,
+        bins=bins,
+        range=[bins[0], bins[-1]],
+        density=density,
+        weights=weights if weighted else None,
+    )
+
+    assert bins_d is bins_d2
+    assert_eq(hist_d, hist)
+    assert_eq(bins_d2, bins)
+
+
+def test_histogram_delayed_n_bins_raises_with_density():
+    data = da.random.random(10, chunks=2)
+    with pytest.raises(
+        NotImplementedError, match="`bins` cannot be a scalar Dask object"
+    ):
+        da.histogram(data, bins=da.array(10), range=[0, 1], density=True)
 
 
 def test_cov():
@@ -810,9 +932,6 @@ def test_shape(shape):
 )
 @pytest.mark.parametrize("reverse", [True, False])
 def test_union1d(shape, reverse):
-    if any(len(x) > 1 for x in shape) and not _numpy_115:
-        pytest.skip("NumPy-10563.")
-
     s1, s2 = shape
     x1 = np.arange(12).reshape(s1)
     x2 = np.arange(6, 18).reshape(s2)
@@ -1135,6 +1254,9 @@ def test_argwhere_obj():
 
 
 def test_argwhere_str():
+    # We may have behavior differences with NumPy for strings
+    # with just spaces, depending on the version of NumPy.
+    # https://github.com/numpy/numpy/issues/9875
     x = np.array(list("Hello world"))
     d = da.from_array(x, chunks=(4,))
 
@@ -1288,7 +1410,10 @@ def test_count_nonzero_obj_axis(axis):
 
 
 def test_count_nonzero_str():
-    x = np.array(list("Hello world"))
+    # We may have behavior differences with NumPy for strings
+    # with just spaces, depending on the version of NumPy.
+    # https://github.com/numpy/numpy/issues/9875
+    x = np.array(list("Hellow orld"))
     d = da.from_array(x, chunks=(4,))
 
     x_c = np.count_nonzero(x)
@@ -1338,10 +1463,6 @@ def test_nonzero_method():
             assert_eq(d_nz[i], x_nz[i])
 
 
-@pytest.mark.skipif(
-    LooseVersion(np.__version__) < LooseVersion("1.14.0"),
-    reason="NumPy 1.14.0+ needed for `unravel_index` to take an empty shape.",
-)
 def test_unravel_index_empty():
     shape = tuple()
     findices = np.array(0, dtype=int)
@@ -1381,6 +1502,28 @@ def test_unravel_index():
         assert_eq(darr.vindex[d_indices], arr[indices])
 
 
+@pytest.mark.parametrize(
+    "arr, chunks, kwargs",
+    [
+        # Numpy doctests:
+        ([[3, 6, 6], [4, 5, 1]], (2, 3), dict(dims=(7, 6), order="C")),
+        ([[3, 6, 6], [4, 5, 1]], (2, 1), dict(dims=(7, 6), order="F")),
+        ([[3, 6, 6], [4, 5, 1]], 1, dict(dims=(4, 6), mode="clip")),
+        ([[3, 6, 6], [4, 5, 1]], (2, 3), dict(dims=(4, 4), mode=("clip", "wrap"))),
+        # Shape tests:
+        ([[3, 6, 6]], (1, 1), dict(dims=(7), order="C")),
+        ([[3, 6, 6], [4, 5, 1], [8, 6, 2]], (3, 1), dict(dims=(7, 6, 9), order="C")),
+    ],
+)
+def test_ravel_multi_index(arr, chunks, kwargs):
+    arr = np.asarray(arr)
+    darr = da.from_array(arr, chunks=chunks)
+    assert_eq(
+        da.ravel_multi_index(darr, **kwargs).compute(),
+        np.ravel_multi_index(arr, **kwargs),
+    )
+
+
 def test_coarsen():
     x = np.random.randint(10, size=(24, 24))
     d = da.from_array(x, chunks=(4, 8))
@@ -1399,13 +1542,31 @@ def test_coarsen():
 
 def test_coarsen_with_excess():
     x = da.arange(10, chunks=5)
-    assert_eq(da.coarsen(np.min, x, {0: 3}, trim_excess=True), np.array([0, 5]))
+    assert_eq(da.coarsen(np.min, x, {0: 5}, trim_excess=True), np.array([0, 5]))
     assert_eq(
         da.coarsen(np.sum, x, {0: 3}, trim_excess=True),
-        np.array([0 + 1 + 2, 5 + 6 + 7]),
+        np.array([0 + 1 + 2, 3 + 4 + 5, 6 + 7 + 8]),
     )
 
 
+def test_coarsen_bad_chunks():
+
+    x1 = da.arange(10, chunks=5)
+    x2 = x1.rechunk((1, 2, 3, 4))
+    assert_eq(da.coarsen(np.sum, x1, {0: 5}), da.coarsen(np.sum, x2, {0: 5}))
+
+
+def test_aligned_coarsen_chunks():
+
+    from ..routines import aligned_coarsen_chunks as acc
+
+    assert acc((20, 10, 15, 23, 24), 10) == (20, 10, 20, 20, 20, 2)
+    assert acc((20, 10, 15, 42, 23, 24), 10) == (20, 10, 20, 40, 20, 20, 4)
+    assert acc((20, 10, 15, 47, 23, 24), 10) == (20, 10, 20, 50, 20, 10, 9)
+    assert acc((2, 10, 15, 47, 23, 24), 10) == (10, 20, 50, 20, 20, 1)
+    assert acc((10, 20, 30, 40, 2), 10) == (10, 20, 30, 40, 2)
+
+
 def test_insert():
     x = np.random.randint(10, size=(10, 10))
     a = da.from_array(x, chunks=(5, 5))
@@ -1715,3 +1876,11 @@ def test_average_raises():
 
     with pytest.warns(RuntimeWarning):
         da.average(d_a, weights=da.zeros_like(d_a)).compute()
+
+
+def test_iscomplexobj():
+    a = da.from_array(np.array([1, 2]), 2)
+    assert np.iscomplexobj(a) is False
+
+    a = da.from_array(np.array([1, 2 + 0j]), 2)
+    assert np.iscomplexobj(a) is True
