@@ -1,23 +1,25 @@
 import itertools
 import warnings
-from collections.abc import Mapping
 
 import numpy as np
 
-try:
-    import cytoolz as toolz
-except ImportError:
-    import toolz
+import tlz as toolz
 
-from .core import reverse_dict
-from .delayed import to_task_dask
-from .highlevelgraph import HighLevelGraph
+from .core import reverse_dict, flatten, keys_in_tasks
+from .delayed import unpack_collections
+from .highlevelgraph import HighLevelGraph, Layer
 from .optimization import SubgraphCallable, fuse
-from .utils import ensure_dict, homogeneous_deepmap, apply
+from .utils import (
+    ensure_dict,
+    homogeneous_deepmap,
+    apply,
+    stringify,
+    stringify_collection_keys,
+)
 
 
 def subs(task, substitution):
-    """ Create a new task with the values substituted
+    """Create a new task with the values substituted
 
     This is like dask.core.subs, but takes a dict of many substitutions to
     perform simultaneously.  It is not as concerned with micro performance.
@@ -55,7 +57,7 @@ def blockwise(
     dependencies=(),
     **kwargs
 ):
-    """ Create a Blockwise symbolic mutable mapping
+    """Create a Blockwise symbolic mutable mapping
 
     This is like the ``make_blockwise_graph`` function, but rather than construct a dict, it
     returns a symbolic Blockwise object.
@@ -76,16 +78,26 @@ def blockwise(
     } | set(output_indices)
     sub = {k: blockwise_token(i, ".") for i, k in enumerate(sorted(unique_indices))}
     output_indices = index_subs(tuple(output_indices), sub)
-    arrind_pairs[1::2] = [tuple(a) if a is not None else a for a in arrind_pairs[1::2]]
-    arrind_pairs[1::2] = [index_subs(a, sub) for a in arrind_pairs[1::2]]
+    a_pairs_list = []
+    for a in arrind_pairs[1::2]:
+        if a is not None:
+            val = tuple(a)
+        else:
+            val = a
+        a_pairs_list.append(index_subs(val, sub))
+
+    arrind_pairs[1::2] = a_pairs_list
     new_axes = {index_subs((k,), sub)[0]: v for k, v in new_axes.items()}
 
     # Unpack dask values in non-array arguments
-    argpairs = list(toolz.partition(2, arrind_pairs))
+    argpairs = toolz.partition(2, arrind_pairs)
 
     # separate argpairs into two separate tuples
-    inputs = tuple([name for name, _ in argpairs])
-    inputs_indices = tuple([index for _, index in argpairs])
+    inputs = []
+    inputs_indices = []
+    for name, index in argpairs:
+        inputs.append(name)
+        inputs_indices.append(index)
 
     # Unpack delayed objects in kwargs
     new_keys = {n for c in dependencies for n in c.__dask_layers__()}
@@ -95,16 +107,16 @@ def blockwise(
             blockwise_token(i) for i in range(len(inputs), len(inputs) + len(new_keys))
         )
         sub = dict(zip(new_keys, new_tokens))
-        inputs = inputs + tuple(new_keys)
-        inputs_indices = inputs_indices + (None,) * len(new_keys)
+        inputs.extend(new_keys)
+        inputs_indices.extend((None,) * len(new_keys))
         kwargs = subs(kwargs, sub)
 
     indices = [(k, v) for k, v in zip(inputs, inputs_indices)]
-    keys = tuple(map(blockwise_token, range(len(inputs))))
+    keys = map(blockwise_token, range(len(inputs)))
 
     # Construct local graph
     if not kwargs:
-        subgraph = {output: (func,) + keys}
+        subgraph = {output: (func,) + tuple(keys)}
     else:
         _keys = list(keys)
         if new_keys:
@@ -125,8 +137,8 @@ def blockwise(
     return subgraph
 
 
-class Blockwise(Mapping):
-    """ Tensor Operation
+class Blockwise(Layer):
+    """Tensor Operation
 
     This is a lazily constructed mapping for tensor operation graphs.
     This defines a dictionary using an operation and an indexing pattern.
@@ -156,10 +168,17 @@ class Blockwise(Mapping):
         single input to the block function
     new_axes: Dict
         New index dimensions that may have been created, and their extent
-
+    output_blocks: Set[Tuple]
+        Specify a specific set of required output blocks. Since the graph
+        will only contain the necessary tasks to generate these outputs,
+        this kwarg can be used to "cull" the abstract layer (without needing
+        to materialize the low-level graph).
+    annotations: dict (optional)
+        Layer annotations
 
     See Also
     --------
+    dask.blockwise.BlockwiseIO
     dask.blockwise.blockwise
     dask.array.blockwise
     """
@@ -173,37 +192,78 @@ class Blockwise(Mapping):
         numblocks,
         concatenate=None,
         new_axes=None,
+        output_blocks=None,
+        annotations=None,
     ):
+        super().__init__(annotations=annotations)
         self.output = output
         self.output_indices = tuple(output_indices)
+        self.output_blocks = output_blocks
         self.dsk = dsk
         self.indices = tuple(
             (name, tuple(ind) if ind is not None else ind) for name, ind in indices
         )
         self.numblocks = numblocks
-        self.concatenate = concatenate
+        # optimize_blockwise won't merge where `concatenate` doesn't match, so
+        # enforce a canonical value if there are no axes for reduction.
+        if all(
+            all(i in output_indices for i in ind)
+            for name, ind in indices
+            if ind is not None
+        ):
+            self.concatenate = None
+        else:
+            self.concatenate = concatenate
         self.new_axes = new_axes or {}
 
+    @property
+    def dims(self):
+        """Returns a dictionary mapping between each index specified in
+        `self.indices` and the number of output blocks for that indice.
+        """
+        if not hasattr(self, "_dims"):
+            self._dims = _make_dims(self.indices, self.numblocks, self.new_axes)
+        return self._dims
+
     def __repr__(self):
         return "Blockwise<{} -> {}>".format(self.indices, self.output)
 
     @property
     def _dict(self):
         if hasattr(self, "_cached_dict"):
-            return self._cached_dict
+            return self._cached_dict["dsk"]
         else:
             keys = tuple(map(blockwise_token, range(len(self.indices))))
-            func = SubgraphCallable(self.dsk, self.output, keys)
-            self._cached_dict = make_blockwise_graph(
+            dsk, _ = fuse(self.dsk, [self.output])
+            func = SubgraphCallable(dsk, self.output, keys)
+
+            dsk = make_blockwise_graph(
                 func,
                 self.output,
                 self.output_indices,
                 *list(toolz.concat(self.indices)),
                 new_axes=self.new_axes,
                 numblocks=self.numblocks,
-                concatenate=self.concatenate
+                concatenate=self.concatenate,
+                output_blocks=self.output_blocks,
+                dims=self.dims,
             )
-        return self._cached_dict
+
+            self._cached_dict = {"dsk": dsk}
+        return self._cached_dict["dsk"]
+
+    def get_output_keys(self):
+        if self.output_blocks:
+            # Culling has already generated a list of output blocks
+            return {(self.output, *p) for p in self.output_blocks}
+
+        # Return all possible output keys (no culling)
+        return {
+            (self.output, *p)
+            for p in itertools.product(
+                *[range(self.dims[i]) for i in self.output_indices]
+            )
+        }
 
     def __getitem__(self, key):
         return self._dict[key]
@@ -216,16 +276,268 @@ class Blockwise(Mapping):
 
     def _out_numblocks(self):
         d = {}
+        out_d = {}
         indices = {k: v for k, v in self.indices if v is not None}
         for k, v in self.numblocks.items():
             for a, b in zip(indices[k], v):
                 d[a] = max(d.get(a, 0), b)
+                if a in self.output_indices:
+                    out_d[a] = d[a]
 
-        return {k: v for k, v in d.items() if k in self.output_indices}
+        return out_d
+
+    def is_materialized(self):
+        return hasattr(self, "_cached_dict")
+
+    def __dask_distributed_pack__(self, client):
+        from distributed.worker import dumps_function
+        from distributed.utils import CancelledError
+        from distributed.utils_comm import unpack_remotedata
+
+        keys = tuple(map(blockwise_token, range(len(self.indices))))
+        dsk, _ = fuse(self.dsk, [self.output])
+
+        dsk = (SubgraphCallable(dsk, self.output, keys),)
+        dsk, dsk_unpacked_futures = unpack_remotedata(dsk, byte_keys=True)
+
+        func = dumps_function(dsk[0])
+        func_future_args = dsk[1:]
+
+        indices = list(toolz.concat(self.indices))
+        indices, indices_unpacked_futures = unpack_remotedata(indices, byte_keys=True)
+
+        # Check the legality of the unpacked futures
+        for future in itertools.chain(dsk_unpacked_futures, indices_unpacked_futures):
+            if future.client is not client:
+                raise ValueError(
+                    "Inputs contain futures that were created by another client."
+                )
+            if stringify(future.key) not in client.futures:
+                raise CancelledError(stringify(future.key))
+
+        # All blockwise tasks will depend on the futures in `indices`
+        global_dependencies = tuple(stringify(f.key) for f in indices_unpacked_futures)
+
+        ret = {
+            "output": self.output,
+            "output_indices": self.output_indices,
+            "func": func,
+            "func_future_args": func_future_args,
+            "global_dependencies": global_dependencies,
+            "indices": indices,
+            "numblocks": self.numblocks,
+            "concatenate": self.concatenate,
+            "new_axes": self.new_axes,
+            "annotations": self.pack_annotations(),
+            "output_blocks": self.output_blocks,
+            "dims": self.dims,
+        }
+
+        return ret
+
+    @classmethod
+    def __dask_distributed_unpack__(cls, state, dsk, dependencies, annotations):
+        raw, raw_deps = make_blockwise_graph(
+            state["func"],
+            state["output"],
+            state["output_indices"],
+            *state["indices"],
+            new_axes=state["new_axes"],
+            numblocks=state["numblocks"],
+            concatenate=state["concatenate"],
+            output_blocks=state["output_blocks"],
+            dims=state["dims"],
+            return_key_deps=True,
+            deserializing=True,
+            func_future_args=state["func_future_args"],
+        )
+        global_dependencies = list(state["global_dependencies"])
+
+        if state["annotations"]:
+            annotations.update(cls.expand_annotations(state["annotations"], raw.keys()))
+
+        raw = {stringify(k): stringify_collection_keys(v) for k, v in raw.items()}
+        dsk.update(raw)
+
+        for k, v in raw_deps.items():
+            dependencies[stringify(k)] = [stringify(d) for d in v] + global_dependencies
+
+    def _cull_dependencies(self, all_hlg_keys, output_blocks):
+        """Determine the necessary dependencies to produce `output_blocks`.
+
+        This method does not require graph materialization.
+        """
+
+        # Check `concatenate` option
+        concatenate = None
+        if self.concatenate is True:
+            from dask.array.core import concatenate_axes as concatenate
+
+        # Generate coordinate map
+        (coord_maps, concat_axes, dummies) = _get_coord_mapping(
+            self.dims,
+            self.output,
+            self.output_indices,
+            self.numblocks,
+            self.indices,
+            concatenate,
+        )
+
+        # Gather constant dependencies (for all output keys)
+        const_deps = set()
+        for (arg, ind) in self.indices:
+            if ind is None and isinstance(arg, str):
+                if arg in all_hlg_keys:
+                    const_deps.add(arg)
+
+        # Get dependencies for each output block
+        key_deps = {}
+        for out_coords in output_blocks:
+            deps = set()
+            coords = out_coords + dummies
+            for cmap, axes, (arg, ind) in zip(coord_maps, concat_axes, self.indices):
+                if ind is not None:
+                    arg_coords = tuple(coords[c] for c in cmap)
+                    if axes:
+                        tups = lol_product((arg,), arg_coords)
+                        deps.update(flatten(tups))
+                        if concatenate:
+                            tups = (concatenate, tups, axes)
+                    else:
+                        tups = (arg,) + arg_coords
+                        deps.add(tups)
+            key_deps[(self.output,) + out_coords] = deps | const_deps
+
+        return key_deps
+
+    def _cull(self, output_blocks):
+        return Blockwise(
+            self.output,
+            self.output_indices,
+            self.dsk,
+            self.indices,
+            self.numblocks,
+            concatenate=self.concatenate,
+            new_axes=self.new_axes,
+            output_blocks=output_blocks,
+            annotations=self.annotations,
+        )
+
+    def cull(self, keys, all_hlg_keys):
+
+        # Culling is simple for Blockwise layers.  We can just
+        # collect a set of required output blocks (tuples), and
+        # only construct graph for these blocks in `make_blockwise_graph`
+
+        output_blocks = set()
+        for key in keys:
+            if key[0] == self.output:
+                output_blocks.add(key[1:])
+        culled_deps = self._cull_dependencies(all_hlg_keys, output_blocks)
+        out_size = tuple([self.dims[i] for i in self.output_indices])
+        if np.prod(out_size) != len(culled_deps):
+            culled_layer = self._cull(output_blocks)
+            return culled_layer, culled_deps
+        else:
+            return self, culled_deps
+
+
+def _get_coord_mapping(
+    dims,
+    output,
+    out_indices,
+    numblocks,
+    argpairs,
+    concatenate,
+):
+    """Calculate coordinate mapping for graph construction.
+
+    This function handles the high-level logic behind Blockwise graph
+    construction. The output is a tuple containing: The mapping between
+    input and output block coordinates (`coord_maps`), the axes along
+    which to concatenate for each input (`concat_axes`), and the dummy
+    indices needed for broadcasting (`dummies`).
+
+    Used by `make_blockwise_graph` and `Blockwise._cull_dependencies`.
+
+    Parameters
+    ----------
+    dims: dict
+        Mapping between each index specified in `argpairs` and
+        the number of output blocks for that index. Corresponds
+        to the Blockwise `dims` attribute.
+    output: str
+        Corresponds to the Blockwise `output` attribute.
+    out_indices: tuple
+        Corresponds to the Blockwise `output_indices` attribute.
+    numblocks: dict
+        Corresponds to the Blockwise `numblocks` attribute.
+    argpairs: tuple
+        Corresponds to the Blockwise `indices` attribute.
+    concatenate: bool
+        Corresponds to the Blockwise `concatenate` attribute.
+    """
+
+    block_names = set()
+    all_indices = set()
+    for name, ind in argpairs:
+        if ind is not None:
+            block_names.add(name)
+            for x in ind:
+                all_indices.add(x)
+    assert set(numblocks) == block_names
+
+    dummy_indices = all_indices - set(out_indices)
+
+    # For each position in the output space, we'll construct a
+    # "coordinate set" that consists of
+    # - the output indices
+    # - the dummy indices
+    # - the dummy indices, with indices replaced by zeros (for broadcasting), we
+    #   are careful to only emit a single dummy zero when concatenate=True to not
+    #   concatenate the same array with itself several times.
+    # - a 0 to assist with broadcasting.
+
+    index_pos, zero_pos = {}, {}
+    for i, ind in enumerate(out_indices):
+        index_pos[ind] = i
+        zero_pos[ind] = -1
+
+    _dummies_list = []
+    for i, ind in enumerate(dummy_indices):
+        index_pos[ind] = 2 * i + len(out_indices)
+        zero_pos[ind] = 2 * i + 1 + len(out_indices)
+        reps = 1 if concatenate else dims[ind]
+        _dummies_list.append([list(range(dims[ind])), [0] * reps])
+
+    # ([0, 1, 2], [0, 0, 0], ...)  For a dummy index of dimension 3
+    dummies = tuple(itertools.chain.from_iterable(_dummies_list))
+    dummies += (0,)
+
+    # For each coordinate position in each input, gives the position in
+    # the coordinate set.
+    coord_maps = []
+
+    # Axes along which to concatenate, for each input
+    concat_axes = []
+    for arg, ind in argpairs:
+        if ind is not None:
+            coord_maps.append(
+                [
+                    zero_pos[i] if nb == 1 else index_pos[i]
+                    for i, nb in zip(ind, numblocks[arg])
+                ]
+            )
+            concat_axes.append([n for n, i in enumerate(ind) if i in dummy_indices])
+        else:
+            coord_maps.append(None)
+            concat_axes.append(None)
+
+    return coord_maps, concat_axes, dummies
 
 
 def make_blockwise_graph(func, output, out_indices, *arrind_pairs, **kwargs):
-    """ Tensor operation
+    """Tensor operation
 
     Applies a function, ``func``, across blocks from many different input
     collections.  We arrange the pattern with which those blocks interact with
@@ -330,78 +642,138 @@ def make_blockwise_graph(func, output, out_indices, *arrind_pairs, **kwargs):
     dask.array.blockwise
     dask.blockwise.blockwise
     """
+
     numblocks = kwargs.pop("numblocks")
     concatenate = kwargs.pop("concatenate", None)
     new_axes = kwargs.pop("new_axes", {})
+    output_blocks = kwargs.pop("output_blocks", None)
+    dims = kwargs.pop("dims", None)
     argpairs = list(toolz.partition(2, arrind_pairs))
 
+    deserializing = kwargs.pop("deserializing", False)
+    func_future_args = kwargs.pop("func_future_args", None)
+    return_key_deps = kwargs.pop("return_key_deps", False)
+    if return_key_deps:
+        key_deps = {}
+
+    if deserializing:
+        from distributed.worker import warn_dumps, dumps_function
+
     if concatenate is True:
         from dask.array.core import concatenate_axes as concatenate
 
-    assert set(numblocks) == {name for name, ind in argpairs if ind is not None}
-
-    all_indices = {x for _, ind in argpairs if ind for x in ind}
-    dummy_indices = all_indices - set(out_indices)
-
     # Dictionary mapping {i: 3, j: 4, ...} for i, j, ... the dimensions
-    dims = broadcast_dimensions(argpairs, numblocks)
-    for k, v in new_axes.items():
-        dims[k] = len(v) if isinstance(v, tuple) else 1
+    dims = dims or _make_dims(argpairs, numblocks, new_axes)
 
-    # (0, 0), (0, 1), (0, 2), (1, 0), ...
-    keytups = list(itertools.product(*[range(dims[i]) for i in out_indices]))
-    # {i: 0, j: 0}, {i: 0, j: 1}, ...
-    keydicts = [dict(zip(out_indices, tup)) for tup in keytups]
-
-    # {j: [1, 2, 3], ...}  For j a dummy index of dimension 3
-    dummies = dict((i, list(range(dims[i]))) for i in dummy_indices)
-
-    dsk = {}
-
-    # Create argument lists
-    valtups = []
-    for kd in keydicts:
-        args = []
-        for arg, ind in argpairs:
-            if ind is None:
-                args.append(arg)
-            else:
-                tups = lol_tuples((arg,), ind, kd, dummies)
-                if any(nb == 1 for nb in numblocks[arg]):
-                    tups2 = zero_broadcast_dimensions(tups, numblocks[arg])
-                else:
-                    tups2 = tups
-                if concatenate and isinstance(tups2, list):
-                    axes = [n for n, i in enumerate(ind) if i in dummies]
-                    tups2 = (concatenate, tups2, axes)
-                args.append(tups2)
-        valtups.append(args)
-
-    if not kwargs:  # will not be used in an apply, should be a tuple
-        valtups = [tuple(vt) for vt in valtups]
-
-    # Add heads to tuples
-    keys = [(output,) + kt for kt in keytups]
+    # Generate the abstract "plan" before constructing
+    # the actual graph
+    (coord_maps, concat_axes, dummies) = _get_coord_mapping(
+        dims,
+        output,
+        out_indices,
+        numblocks,
+        argpairs,
+        concatenate,
+    )
 
     # Unpack delayed objects in kwargs
+    dsk2 = {}
     if kwargs:
-        task, dsk2 = to_task_dask(kwargs)
+        task, dsk2 = unpack_collections(kwargs)
         if dsk2:
-            dsk.update(ensure_dict(dsk2))
             kwargs2 = task
         else:
             kwargs2 = kwargs
-        vals = [(apply, func, vt, kwargs2) for vt in valtups]
+
+    # Apply Culling.
+    # Only need to construct the specified set of output blocks
+    output_blocks = output_blocks or itertools.product(
+        *[range(dims[i]) for i in out_indices]
+    )
+
+    dsk = {}
+    # Create argument lists
+    for out_coords in output_blocks:
+        deps = set()
+        coords = out_coords + dummies
+        args = []
+        for cmap, axes, (arg, ind) in zip(coord_maps, concat_axes, argpairs):
+            if ind is None:
+                args.append(arg)
+            else:
+                arg_coords = tuple(coords[c] for c in cmap)
+                if axes:
+                    tups = lol_product((arg,), arg_coords)
+                    deps.update(flatten(tups))
+
+                    if concatenate:
+                        tups = (concatenate, tups, axes)
+                else:
+                    tups = (arg,) + arg_coords
+                    deps.add(tups)
+                args.append(tups)
+        out_key = (output,) + out_coords
+
+        if deserializing:
+            deps.update(func_future_args)
+            args = stringify_collection_keys(args) + list(func_future_args)
+            if kwargs:
+                val = {
+                    "function": dumps_function(apply),
+                    "args": warn_dumps(args),
+                    "kwargs": warn_dumps(kwargs2),
+                }
+            else:
+                val = {"function": func, "args": warn_dumps(args)}
+        else:
+            if kwargs:
+                val = (apply, func, args, kwargs2)
+            else:
+                args.insert(0, func)
+                val = tuple(args)
+        dsk[out_key] = val
+        if return_key_deps:
+            key_deps[out_key] = deps
+
+    if dsk2:
+        dsk.update(ensure_dict(dsk2))
+
+    if return_key_deps:
+        return dsk, key_deps
     else:
-        vals = [(func,) + vt for vt in valtups]
+        return dsk
 
-    dsk.update(dict(zip(keys, vals)))
 
-    return dsk
+def lol_product(head, values):
+    """List of list of tuple keys, similar to `itertools.product`.
+
+    Parameters
+    ----------
+
+    head : tuple
+        Prefix prepended to all results.
+    values : sequence
+        Mix of singletons and lists. Each list is substituted with every
+        possible value and introduces another level of list in the output.
+    Examples
+    --------
+
+    >>> lol_product(('x',), (1, 2, 3))
+    ('x', 1, 2, 3)
+    >>> lol_product(('x',), (1, [2, 3], 4, [5, 6]))  # doctest: +NORMALIZE_WHITESPACE
+    [[('x', 1, 2, 4, 5), ('x', 1, 2, 4, 6)],
+     [('x', 1, 3, 4, 5), ('x', 1, 3, 4, 6)]]
+    """
+    if not values:
+        return head
+    elif isinstance(values[0], list):
+        return [lol_product(head + (x,), values[1:]) for x in values[0]]
+    else:
+        return lol_product(head + (values[0],), values[1:])
 
 
 def lol_tuples(head, ind, values, dummies):
-    """ List of list of tuple keys
+    """List of list of tuple keys
 
     Parameters
     ----------
@@ -424,9 +796,6 @@ def lol_tuples(head, ind, values, dummies):
     >>> lol_tuples(('x',), 'ij', {'i': 1}, {'j': range(3)})
     [('x', 1, 0), ('x', 1, 1), ('x', 1, 2)]
 
-    >>> lol_tuples(('x',), 'ij', {'i': 1}, {'j': range(3)})
-    [('x', 1, 0), ('x', 1, 1), ('x', 1, 2)]
-
     >>> lol_tuples(('x',), 'ijk', {'i': 1}, {'j': [0, 1, 2], 'k': [0, 1]}) # doctest: +NORMALIZE_WHITESPACE
     [[('x', 1, 0, 0), ('x', 1, 0, 1)],
      [('x', 1, 1, 0), ('x', 1, 1, 1)],
@@ -443,7 +812,7 @@ def lol_tuples(head, ind, values, dummies):
 
 
 def optimize_blockwise(graph, keys=()):
-    """ High level optimization of stacked Blockwise layers
+    """High level optimization of stacked Blockwise layers
 
     For operations that have multiple Blockwise operations one after the other, like
     ``x.T + 123`` we can fuse these into a single Blockwise operation.  This happens
@@ -513,6 +882,7 @@ def _optimize_blockwise(full_graph, keys=()):
             deps = set(blockwise_layers)
             while deps:  # we gather as many sub-layers as we can
                 dep = deps.pop()
+
                 if dep not in layers:
                     stack.append(dep)
                     continue
@@ -531,6 +901,13 @@ def _optimize_blockwise(full_graph, keys=()):
                 ):
                     stack.append(dep)
                     continue
+                if (
+                    blockwise_layers
+                    and layers[next(iter(blockwise_layers))].annotations
+                    != layers[dep].annotations
+                ):
+                    stack.append(dep)
+                    continue
 
                 # passed everything, proceed
                 blockwise_layers.add(dep)
@@ -553,7 +930,14 @@ def _optimize_blockwise(full_graph, keys=()):
             # Merge these Blockwise layers into one
             new_layer = rewrite_blockwise([layers[l] for l in blockwise_layers])
             out[layer] = new_layer
-            dependencies[layer] = {k for k, v in new_layer.indices if v is not None}
+
+            new_deps = set()
+            for k, v in new_layer.indices:
+                if v is None:
+                    new_deps |= keys_in_tasks(full_graph.dependencies, [k])
+                else:
+                    new_deps.add(k)
+            dependencies[layer] = new_deps
         else:
             out[layer] = layers[layer]
             dependencies[layer] = full_graph.dependencies.get(layer, set())
@@ -563,7 +947,7 @@ def _optimize_blockwise(full_graph, keys=()):
 
 
 def rewrite_blockwise(inputs):
-    """ Rewrite a stack of Blockwise expressions into a single blockwise expression
+    """Rewrite a stack of Blockwise expressions into a single blockwise expression
 
     Given a set of Blockwise layers, combine them into a single layer.  The provided
     layers are expected to fit well together.  That job is handled by
@@ -687,7 +1071,7 @@ def rewrite_blockwise(inputs):
     numblocks = toolz.merge([inp.numblocks for inp in inputs.values()])
     numblocks = {k: v for k, v in numblocks.items() if v is None or k in indices_check}
 
-    out = Blockwise(
+    return Blockwise(
         root,
         inputs[root].output_indices,
         dsk,
@@ -695,10 +1079,9 @@ def rewrite_blockwise(inputs):
         numblocks=numblocks,
         new_axes=new_axes,
         concatenate=concatenate,
+        annotations=inputs[root].annotations,
     )
 
-    return out
-
 
 def zero_broadcast_dimensions(lol, nblocks):
     """
@@ -723,7 +1106,7 @@ def zero_broadcast_dimensions(lol, nblocks):
 
 
 def broadcast_dimensions(argpairs, numblocks, sentinels=(1, (1,)), consolidate=None):
-    """ Find block dimensions from arguments
+    """Find block dimensions from arguments
 
     Parameters
     ----------
@@ -782,6 +1165,16 @@ def broadcast_dimensions(argpairs, numblocks, sentinels=(1, (1,)), consolidate=N
     return toolz.valmap(toolz.first, g2)
 
 
+def _make_dims(indices, numblocks, new_axes):
+    """Returns a dictionary mapping between each index specified in
+    `indices` and the number of output blocks for that indice.
+    """
+    dims = broadcast_dimensions(indices, numblocks)
+    for k, v in new_axes.items():
+        dims[k] = len(v) if isinstance(v, tuple) else 1
+    return dims
+
+
 def fuse_roots(graph: HighLevelGraph, keys: list):
     """
     Fuse nearby layers if they don't have dependencies
@@ -807,7 +1200,7 @@ def fuse_roots(graph: HighLevelGraph, keys: list):
     graph: HighLevelGraph
         The full graph of the computation
     keys: list
-        The output keys of the comptuation, to be passed on to fuse
+        The output keys of the computation, to be passed on to fuse
 
     See Also
     --------
@@ -825,12 +1218,14 @@ def fuse_roots(graph: HighLevelGraph, keys: list):
             and len(deps) > 1
             and not any(dependencies[dep] for dep in deps)  # no need to fuse if 0 or 1
             and all(len(dependents[dep]) == 1 for dep in deps)
+            and all(layer.annotations == graph.layers[dep].annotations for dep in deps)
         ):
             new = toolz.merge(layer, *[layers[dep] for dep in deps])
             new, _ = fuse(new, keys, ave_width=len(deps))
 
             for dep in deps:
                 del layers[dep]
+                del dependencies[dep]
 
             layers[name] = new
             dependencies[name] = set()
