@@ -1,14 +1,14 @@
 import ast
+import base64
 import builtins
+from collections.abc import Mapping
+import json
 import os
 import sys
 import threading
-from collections.abc import Mapping
+import warnings
 
-try:
-    import yaml
-except ImportError:
-    yaml = None
+import yaml
 
 
 no_default = "__no_default__"
@@ -61,7 +61,7 @@ def canonical_name(k, config):
 
 
 def update(old, new, priority="new"):
-    """ Update a nested dictionary with values from another
+    """Update a nested dictionary with values from another
 
     This is like dict.update except that it smoothly merges nested values
 
@@ -104,7 +104,7 @@ def update(old, new, priority="new"):
 
 
 def merge(*dicts):
-    """ Update a sequence of nested dictionaries
+    """Update a sequence of nested dictionaries
 
     This prefers the values in the latter dictionaries to those in the former
 
@@ -126,7 +126,7 @@ def merge(*dicts):
 
 
 def collect_yaml(paths=paths):
-    """ Collect configuration from yaml files
+    """Collect configuration from yaml files
 
     This searches through a list of paths, expands to find all yaml or json
     files, and then parses each file.
@@ -169,7 +169,7 @@ def collect_yaml(paths=paths):
 
 
 def collect_env(env=None):
-    """ Collect config from environment variables
+    """Collect config from environment variables
 
     This grabs environment variables of the form "DASK_FOO__BAR_BAZ=123" and
     turns these into config variables of the form ``{"foo": {"bar-baz": 123}}``
@@ -178,10 +178,19 @@ def collect_env(env=None):
     -  Lower-cases the key text
     -  Treats ``__`` (double-underscore) as nested access
     -  Calls ``ast.literal_eval`` on the value
+
+    Any serialized config passed via ``DASK_INTERNAL_INHERIT_CONFIG`` is also set here.
+
     """
+
     if env is None:
         env = os.environ
-    d = {}
+
+    if "DASK_INTERNAL_INHERIT_CONFIG" in env:
+        d = deserialize(env["DASK_INTERNAL_INHERIT_CONFIG"])
+    else:
+        d = {}
+
     for name, value in env.items():
         if name.startswith("DASK_"):
             varname = name[5:].lower().replace("__", ".")
@@ -258,7 +267,7 @@ def ensure_file(source, destination=None, comment=True):
 
 
 class set(object):
-    """ Temporarily set configuration values within a context manager
+    """Temporarily set configuration values within a context manager
 
     Parameters
     ----------
@@ -300,10 +309,13 @@ class set(object):
 
             if arg is not None:
                 for key, value in arg.items():
+                    key = check_deprecations(key)
                     self._assign(key.split("."), value, config)
             if kwargs:
                 for key, value in kwargs.items():
-                    self._assign(key.split("__"), value, config)
+                    key = key.replace("__", ".")
+                    key = check_deprecations(key)
+                    self._assign(key.split("."), value, config)
 
     def __enter__(self):
         return self.config
@@ -341,6 +353,7 @@ class set(object):
             Whether this operation needs to be recorded to allow for rollback.
         """
         key = canonical_name(keys[0], d)
+
         path = path + (key,)
 
         if len(keys) == 1:
@@ -382,11 +395,8 @@ def collect(paths=paths, env=None):
     """
     if env is None:
         env = os.environ
-    configs = []
-
-    if yaml:
-        configs.extend(collect_yaml(paths=paths))
 
+    configs = collect_yaml(paths=paths)
     configs.append(collect_env(env=env))
 
     return merge(*configs)
@@ -424,10 +434,13 @@ def refresh(config=config, defaults=defaults, **kwargs):
     update(config, collect(**kwargs))
 
 
-def get(key, default=no_default, config=config):
+def get(key, default=no_default, config=config, override_with=None):
     """
     Get elements from global config
 
+    If ``override_with`` is not None this value will be passed straight back.
+    Useful for getting kwarg defaults from Dask config.
+
     Use '.' for nested access
 
     Examples
@@ -442,10 +455,18 @@ def get(key, default=no_default, config=config):
     >>> config.get('foo.x.y', default=123)  # doctest: +SKIP
     123
 
+    >>> config.get('foo.y', override_with=None)  # doctest: +SKIP
+    2
+
+    >>> config.get('foo.y', override_with=3)  # doctest: +SKIP
+    3
+
     See Also
     --------
     dask.config.set
     """
+    if override_with is not None:
+        return override_with
     keys = key.split(".")
     result = config
     for k in keys:
@@ -461,7 +482,7 @@ def get(key, default=no_default, config=config):
 
 
 def rename(aliases, config=config):
-    """ Rename old keys to new keys
+    """Rename old keys to new keys
 
     This helps migrate older configuration versions over time
     """
@@ -480,7 +501,7 @@ def rename(aliases, config=config):
 
 
 def update_defaults(new, config=config, defaults=defaults):
-    """ Add a new set of defaults to the configuration
+    """Add a new set of defaults to the configuration
 
     It does two things:
 
@@ -493,7 +514,7 @@ def update_defaults(new, config=config, defaults=defaults):
 
 
 def expand_environment_variables(config):
-    """ Expand environment variables in a nested config dictionary
+    """Expand environment variables in a nested config dictionary
 
     This function will recursively search through any nested dictionaries
     and/or lists.
@@ -522,15 +543,106 @@ def expand_environment_variables(config):
         return config
 
 
-refresh()
+deprecations = {
+    "fuse_ave_width": "optimization.fuse.ave-width",
+    "fuse_max_height": "optimization.fuse.max-height",
+    "fuse_max_width": "optimization.fuse.max-width",
+    "fuse_subgraphs": "optimization.fuse.subgraphs",
+    "fuse_rename_keys": "optimization.fuse.rename-keys",
+    "fuse_max_depth_new_edges": "optimization.fuse.max-depth-new-edges",
+}
 
 
-if yaml:
+def check_deprecations(key: str, deprecations: dict = deprecations):
+    """Check if the provided value has been renamed or removed
+
+    Parameters
+    ----------
+    key : str
+        The configuration key to check
+    deprecations : Dict[str, str]
+        The mapping of aliases
+
+    Examples
+    --------
+    >>> deprecations = {"old_key": "new_key", "invalid": None}
+    >>> check_deprecations("old_key", deprecations=deprecations)  # doctest: +SKIP
+    UserWarning: Configuration key "old_key" has been deprecated. Please use "new_key" instead.
+
+    >>> check_deprecations("invalid", deprecations=deprecations)
+    Traceback (most recent call last):
+        ...
+    ValueError: Configuration value "invalid" has been removed
+
+    >>> check_deprecations("another_key", deprecations=deprecations)
+    'another_key'
+
+    Returns
+    -------
+    new: str
+        The proper key, whether the original (if no deprecation) or the aliased
+        value
+    """
+    if key in deprecations:
+        new = deprecations[key]
+        if new:
+            warnings.warn(
+                'Configuration key "{}" has been deprecated. '
+                'Please use "{}" instead'.format(key, new)
+            )
+            return new
+        else:
+            raise ValueError('Configuration value "{}" has been removed'.format(key))
+    else:
+        return key
+
+
+def serialize(data):
+    """Serialize config data into a string.
+
+    Typically used to pass config via the ``DASK_INTERNAL_INHERIT_CONFIG`` environment variable.
+
+    Parameters
+    ----------
+    data: json-serializable object
+        The data to serialize
+
+    Returns
+    -------
+    serialized_data: str
+        The serialized data as a string
+
+    """
+    return base64.urlsafe_b64encode(json.dumps(data).encode()).decode()
+
+
+def deserialize(data):
+    """De-serialize config data into the original object.
+
+    Typically when receiving config via the ``DASK_INTERNAL_INHERIT_CONFIG`` environment variable.
+
+    Parameters
+    ----------
+    data: str
+        String serialied by :func:`dask.config.serialize`
+
+    Returns
+    -------
+    deserialized_data: obj
+        The de-serialized data
+
+    """
+    return json.loads(base64.urlsafe_b64decode(data.encode()).decode())
+
+
+def _initialize():
     fn = os.path.join(os.path.dirname(__file__), "dask.yaml")
-    ensure_file(source=fn)
 
     with open(fn) as f:
         _defaults = yaml.safe_load(f)
 
     update_defaults(_defaults)
-    del fn, _defaults
+
+
+refresh()
+_initialize()
