@@ -24,6 +24,8 @@ from .utils import (
     raise_on_meta_error,
     is_series_like,
     is_dataframe_like,
+    PANDAS_GT_100,
+    PANDAS_GT_110,
 )
 from ..base import tokenize
 from ..utils import derived_from, M, funcname, itemgetter
@@ -64,8 +66,7 @@ from ..highlevelgraph import HighLevelGraph
 
 
 def _determine_levels(index):
-    """Determine the correct levels argument to groupby.
-    """
+    """Determine the correct levels argument to groupby."""
     if isinstance(index, (tuple, list)) and len(index) > 1:
         return list(range(len(index)))
     else:
@@ -73,8 +74,7 @@ def _determine_levels(index):
 
 
 def _normalize_index(df, index):
-    """Replace series with column names in an index wherever possible.
-    """
+    """Replace series with column names in an index wherever possible."""
     if not isinstance(df, DataFrame):
         return index
 
@@ -163,24 +163,26 @@ def _groupby_raise_unaligned(df, **kwargs):
 
 
 def _groupby_slice_apply(
-    df, grouper, key, func, *args, group_keys=True, dropna=None, **kwargs
+    df, grouper, key, func, *args, group_keys=True, dropna=None, observed=None, **kwargs
 ):
     # No need to use raise if unaligned here - this is only called after
     # shuffling, which makes everything aligned already
     dropna = {"dropna": dropna} if dropna is not None else {}
-    g = df.groupby(grouper, group_keys=group_keys, **dropna)
+    observed = {"observed": observed} if observed is not None else {}
+    g = df.groupby(grouper, group_keys=group_keys, **observed, **dropna)
     if key:
         g = g[key]
     return g.apply(func, *args, **kwargs)
 
 
 def _groupby_slice_transform(
-    df, grouper, key, func, *args, group_keys=True, dropna=None, **kwargs
+    df, grouper, key, func, *args, group_keys=True, dropna=None, observed=None, **kwargs
 ):
     # No need to use raise if unaligned here - this is only called after
     # shuffling, which makes everything aligned already
     dropna = {"dropna": dropna} if dropna is not None else {}
-    g = df.groupby(grouper, group_keys=group_keys, **dropna)
+    observed = {"observed": observed} if observed is not None else {}
+    g = df.groupby(grouper, group_keys=group_keys, **observed, **dropna)
     if key:
         g = g[key]
 
@@ -275,18 +277,24 @@ class Aggregation(object):
 
 
 def _groupby_aggregate(
-    df, aggfunc=None, levels=None, dropna=None, sort=False, **kwargs
+    df, aggfunc=None, levels=None, dropna=None, sort=False, observed=None, **kwargs
 ):
     dropna = {"dropna": dropna} if dropna is not None else {}
-    return aggfunc(df.groupby(level=levels, sort=sort, **dropna), **kwargs)
+    if not PANDAS_GT_100 and observed:
+        raise NotImplementedError("``observed`` is only supported for pandas >= 1.0.0")
+    observed = {"observed": observed} if observed is not None else {}
+
+    grouped = df.groupby(level=levels, sort=sort, **observed, **dropna)
+    return aggfunc(grouped, **kwargs)
 
 
-def _apply_chunk(df, *index, dropna=None, **kwargs):
+def _apply_chunk(df, *index, dropna=None, observed=None, **kwargs):
     func = kwargs.pop("chunk")
     columns = kwargs.pop("columns")
     dropna = {"dropna": dropna} if dropna is not None else {}
-    g = _groupby_raise_unaligned(df, by=index, **dropna)
+    observed = {"observed": observed} if observed is not None else {}
 
+    g = _groupby_raise_unaligned(df, by=index, **observed, **dropna)
     if is_series_like(df) or columns is None:
         return func(g, **kwargs)
     else:
@@ -402,7 +410,7 @@ def _cov_chunk(df, *index):
     Returns
     -------
     tuple
-        Processed X, Multipled Cols,
+        Processed X, Multiplied Cols,
     """
     if is_series_like(df):
         df = df.to_frame()
@@ -661,7 +669,7 @@ def _build_agg_args(spec):
     Parameters
     ----------
     spec: a list of (result-column, aggregation-function, input-column) triples.
-        To work with all arugment forms understood by pandas use
+        To work with all argument forms understood by pandas use
         ``_normalize_spec`` to normalize the argment before passing it on to
         ``_build_agg_args``.
 
@@ -670,7 +678,7 @@ def _build_agg_args(spec):
     chunk_funcs: a list of (intermediate-column, function, keyword) triples
         that are applied on grouped chunks of the initial dataframe.
 
-    agg_funcs: a list of (intermediate-column, functions, keword) triples that
+    agg_funcs: a list of (intermediate-column, functions, keyword) triples that
         are applied on the grouped concatination of the preprocessed chunks.
 
     finalizers: a list of (result-column, function, keyword) triples that are
@@ -700,8 +708,10 @@ def _build_agg_args(spec):
         impls = _build_agg_args_single(result_column, func, input_column)
 
         # overwrite existing result-columns, generate intermediates only once
-        chunks.update((spec[0], spec) for spec in impls["chunk_funcs"])
-        aggs.update((spec[0], spec) for spec in impls["aggregate_funcs"])
+        for spec in impls["chunk_funcs"]:
+            chunks[spec[0]] = spec
+        for spec in impls["aggregate_funcs"]:
+            aggs[spec[0]] = spec
 
         finalizers.append(impls["finalizer"])
 
@@ -737,6 +747,9 @@ def _build_agg_args_single(result_column, func, input_column):
     elif func == "mean":
         return _build_agg_args_mean(result_column, func, input_column)
 
+    elif func == "list":
+        return _build_agg_args_list(result_column, func, input_column)
+
     elif isinstance(func, Aggregation):
         return _build_agg_args_custom(result_column, func, input_column)
 
@@ -820,6 +833,33 @@ def _build_agg_args_mean(result_column, func, input_column):
     )
 
 
+def _build_agg_args_list(result_column, func, input_column):
+    intermediate = _make_agg_id("list", input_column)
+
+    return dict(
+        chunk_funcs=[
+            (
+                intermediate,
+                _apply_func_to_column,
+                dict(column=input_column, func=lambda s: s.apply(list)),
+            )
+        ],
+        aggregate_funcs=[
+            (
+                intermediate,
+                _apply_func_to_column,
+                dict(
+                    column=intermediate,
+                    func=lambda s0: s0.apply(
+                        lambda chunks: list(it.chain.from_iterable(chunks))
+                    ),
+                ),
+            )
+        ],
+        finalizer=(result_column, itemgetter(intermediate), dict()),
+    )
+
+
 def _build_agg_args_custom(result_column, func, input_column):
     col = _make_agg_id(funcname(func), input_column)
 
@@ -895,18 +935,27 @@ def _groupby_apply_funcs(df, *index, **kwargs):
 
 
 def _compute_sum_of_squares(grouped, column):
-    base = grouped[column] if column is not None else grouped
-    return base.apply(lambda x: (x ** 2).sum())
+    # Note: CuDF cannot use `groupby.apply`.
+    # Need to unpack groupby to compute sum of squares
+    if hasattr(grouped, "grouper"):
+        keys = grouped.grouper
+    else:
+        # Handle CuDF groupby object (different from pandas)
+        keys = grouped.grouping.keys
+    df = grouped.obj[column].pow(2) if column else grouped.obj.pow(2)
+    return df.groupby(keys).sum()
 
 
-def _agg_finalize(df, aggregate_funcs, finalize_funcs, level, sort=False):
+def _agg_finalize(df, aggregate_funcs, finalize_funcs, level, sort=False, **kwargs):
     # finish the final aggregation level
-    df = _groupby_apply_funcs(df, funcs=aggregate_funcs, level=level, sort=sort)
+    df = _groupby_apply_funcs(
+        df, funcs=aggregate_funcs, level=level, sort=sort, **kwargs
+    )
 
     # and finalize the result
     result = collections.OrderedDict()
-    for result_column, func, kwargs in finalize_funcs:
-        result[result_column] = func(df, **kwargs)
+    for result_column, func, finalize_kwargs in finalize_funcs:
+        result[result_column] = func(df, **finalize_kwargs)
 
     return type(df)(result)
 
@@ -974,7 +1023,7 @@ def _cumcount_aggregate(a, b, fill_value=None):
 
 
 class _GroupBy(object):
-    """ Superclass for DataFrameGroupBy and SeriesGroupBy
+    """Superclass for DataFrameGroupBy and SeriesGroupBy
 
     Parameters
     ----------
@@ -992,10 +1041,21 @@ class _GroupBy(object):
     sort: bool, defult None
         Passed along to aggregation methods. If allowed,
         the output aggregation will have sorted keys.
+    observed: bool, default False
+        This only applies if any of the groupers are Categoricals.
+        If True: only show observed values for categorical groupers.
+        If False: show all values for categorical groupers.
     """
 
     def __init__(
-        self, df, by=None, slice=None, group_keys=True, dropna=None, sort=None
+        self,
+        df,
+        by=None,
+        slice=None,
+        group_keys=True,
+        dropna=None,
+        sort=None,
+        observed=None,
     ):
 
         assert isinstance(df, (DataFrame, Series))
@@ -1039,8 +1099,13 @@ class _GroupBy(object):
         if dropna is not None:
             self.dropna["dropna"] = dropna
 
+        # Hold off on setting observed by default: https://github.com/dask/dask/issues/6951
+        self.observed = {}
+        if observed is not None:
+            self.observed["observed"] = observed
+
         self._meta = self.obj._meta.groupby(
-            index_meta, group_keys=group_keys, **self.dropna
+            index_meta, group_keys=group_keys, **self.observed, **self.dropna
         )
 
     @property
@@ -1062,7 +1127,12 @@ class _GroupBy(object):
         else:
             index_meta = self.index
 
-        grouped = sample.groupby(index_meta, group_keys=self.group_keys, **self.dropna)
+        grouped = sample.groupby(
+            index_meta,
+            group_keys=self.group_keys,
+            **self.observed,
+            **self.dropna,
+        )
         return _maybe_slice(grouped, self._slice)
 
     def _aca_agg(
@@ -1090,14 +1160,22 @@ class _GroupBy(object):
             else [self.obj] + self.index,
             chunk=_apply_chunk,
             chunk_kwargs=dict(
-                chunk=func, columns=columns, **chunk_kwargs, **self.dropna
+                chunk=func,
+                columns=columns,
+                **self.observed,
+                **chunk_kwargs,
+                **self.dropna,
             ),
             aggregate=_groupby_aggregate,
             meta=meta,
             token=token,
             split_every=split_every,
             aggregate_kwargs=dict(
-                aggfunc=aggfunc, levels=levels, **aggregate_kwargs, **self.dropna
+                aggfunc=aggfunc,
+                levels=levels,
+                **self.observed,
+                **aggregate_kwargs,
+                **self.dropna,
             ),
             split_out=split_out,
             split_out_setup=split_out_on_index,
@@ -1124,7 +1202,7 @@ class _GroupBy(object):
             columns=columns,
             token=name_part,
             meta=meta,
-            **self.dropna
+            **self.dropna,
         )
 
         cumpart_raw_frame = (
@@ -1153,7 +1231,7 @@ class _GroupBy(object):
             chunk=M.last,
             meta=meta,
             token=name_last,
-            **self.dropna
+            **self.dropna,
         )
 
         # aggregate cumulated partitions and its previous last element
@@ -1517,15 +1595,27 @@ class _GroupBy(object):
         else:
             chunk_args = [self.obj] + self.index
 
+        if not PANDAS_GT_110 and self.dropna:
+            raise NotImplementedError(
+                "dropna is not a valid argument for dask.groupby.agg"
+                f"if pandas < 1.1.0. Pandas version is {pd.__version__}"
+            )
+
         return aca(
             chunk_args,
             chunk=_groupby_apply_funcs,
-            chunk_kwargs=dict(funcs=chunk_funcs),
+            chunk_kwargs=dict(funcs=chunk_funcs, **self.observed, **self.dropna),
             combine=_groupby_apply_funcs,
-            combine_kwargs=dict(funcs=aggregate_funcs, level=levels),
+            combine_kwargs=dict(
+                funcs=aggregate_funcs, level=levels, **self.observed, **self.dropna
+            ),
             aggregate=_agg_finalize,
             aggregate_kwargs=dict(
-                aggregate_funcs=aggregate_funcs, finalize_funcs=finalizers, level=levels
+                aggregate_funcs=aggregate_funcs,
+                finalize_funcs=finalizers,
+                level=levels,
+                **self.observed,
+                **self.dropna,
             ),
             token="aggregate",
             split_every=split_every,
@@ -1536,7 +1626,7 @@ class _GroupBy(object):
 
     @insert_meta_param_description(pad=12)
     def apply(self, func, *args, **kwargs):
-        """ Parallel version of pandas GroupBy.apply
+        """Parallel version of pandas GroupBy.apply
 
         This mimics the pandas version except for the following:
 
@@ -1616,15 +1706,16 @@ class _GroupBy(object):
             token=funcname(func),
             *args,
             group_keys=self.group_keys,
+            **self.observed,
             **self.dropna,
-            **kwargs
+            **kwargs,
         )
 
         return df3
 
     @insert_meta_param_description(pad=12)
     def transform(self, func, *args, **kwargs):
-        """ Parallel version of pandas GroupBy.transform
+        """Parallel version of pandas GroupBy.transform
 
         This mimics the pandas version except for the following:
 
@@ -1704,8 +1795,9 @@ class _GroupBy(object):
             token=funcname(func),
             *args,
             group_keys=self.group_keys,
+            **self.observed,
             **self.dropna,
-            **kwargs
+            **kwargs,
         )
 
         return df3
@@ -1742,16 +1834,14 @@ class DataFrameGroupBy(_GroupBy):
         try:
             return self[key]
         except KeyError as e:
-            raise AttributeError(e)
+            raise AttributeError(e) from e
 
     @derived_from(pd.core.groupby.DataFrameGroupBy)
     def aggregate(self, arg, split_every=None, split_out=1):
         if arg == "size":
             return self.size()
 
-        return super(DataFrameGroupBy, self).aggregate(
-            arg, split_every=split_every, split_out=split_out
-        )
+        return super().aggregate(arg, split_every=split_every, split_out=split_out)
 
     @derived_from(pd.core.groupby.DataFrameGroupBy)
     def agg(self, arg, split_every=None, split_out=1):
@@ -1762,8 +1852,10 @@ class SeriesGroupBy(_GroupBy):
 
     _token_prefix = "series-groupby-"
 
-    def __init__(self, df, by=None, slice=None, **kwargs):
+    def __init__(self, df, by=None, slice=None, observed=None, **kwargs):
         # for any non series object, raise pandas-compat error message
+        # Hold off on setting observed by default: https://github.com/dask/dask/issues/6951
+        observed = {"observed": observed} if observed is not None else {}
 
         if isinstance(df, Series):
             if isinstance(by, Series):
@@ -1774,14 +1866,26 @@ class SeriesGroupBy(_GroupBy):
 
                 non_series_items = [item for item in by if not isinstance(item, Series)]
                 # raise error from pandas, if applicable
-                df._meta.groupby(non_series_items)
+
+                df._meta.groupby(non_series_items, **observed)
             else:
                 # raise error from pandas, if applicable
-                df._meta.groupby(by)
+                df._meta.groupby(by, **observed)
 
-        super(SeriesGroupBy, self).__init__(df, by=by, slice=slice, **kwargs)
+        super().__init__(df, by=by, slice=slice, **observed, **kwargs)
 
+    @derived_from(pd.core.groupby.SeriesGroupBy)
     def nunique(self, split_every=None, split_out=1):
+        """
+        Examples
+        --------
+        >>> import pandas as pd
+        >>> import dask.dataframe as dd
+        >>> d = {'col1': [1, 2, 3, 4], 'col2': [5, 6, 7, 8]}
+        >>> df = pd.DataFrame(data=d)
+        >>> ddf = dd.from_pandas(df, 2)
+        >>> ddf.groupby(['col1']).col2.nunique().compute()
+        """
         name = self._meta.obj.name
         levels = _determine_levels(self.index)
 
@@ -1810,9 +1914,7 @@ class SeriesGroupBy(_GroupBy):
 
     @derived_from(pd.core.groupby.SeriesGroupBy)
     def aggregate(self, arg, split_every=None, split_out=1):
-        result = super(SeriesGroupBy, self).aggregate(
-            arg, split_every=split_every, split_out=split_out
-        )
+        result = super().aggregate(arg, split_every=split_every, split_out=split_out)
         if self._slice:
             result = result[self._slice]
 
