@@ -1,6 +1,5 @@
 import numpy as np
 import pandas as pd
-import pandas.util.testing as tm
 import pytest
 
 import os
@@ -9,6 +8,7 @@ import pathlib
 
 import dask
 import dask.dataframe as dd
+from dask.dataframe._compat import tm
 from dask.utils import tmpfile, tmpdir, dependency_depth
 from dask.dataframe.utils import assert_eq
 
@@ -412,7 +412,7 @@ def test_to_hdf_lock_delays():
     )
     a = dd.from_pandas(df16, 16)
 
-    # adding artifichial delays to make sure last tasks finish first
+    # adding artificial delays to make sure last tasks finish first
     # that's a way to simulate last tasks finishing last
     def delayed_nop(i):
         if i[1] < 10:
@@ -427,7 +427,7 @@ def test_to_hdf_lock_delays():
         assert_eq(df16, out)
 
     # saving to multiple hdf files
-    # adding artifichial delays to make sure last tasks finish first
+    # adding artificial delays to make sure last tasks finish first
     with tmpdir() as dn:
         fn = os.path.join(dn, "data*")
         a = a.apply(delayed_nop, axis=1, meta=a)
@@ -829,3 +829,60 @@ def test_hdf_filenames():
     assert ddf.to_hdf("foo*.hdf5", "key") == ["foo0.hdf5", "foo1.hdf5"]
     os.remove("foo0.hdf5")
     os.remove("foo1.hdf5")
+
+
+def test_hdf_path_exceptions():
+
+    # single file doesn't exist
+    with pytest.raises(IOError):
+        dd.read_hdf("nonexistant_store_X34HJK", "/tmp")
+
+    # a file from a list of files doesn't exist
+    with pytest.raises(IOError):
+        dd.read_hdf(["nonexistant_store_X34HJK", "nonexistant_store_UY56YH"], "/tmp")
+
+    # list of files is empty
+    with pytest.raises(ValueError):
+        dd.read_hdf([], "/tmp")
+
+
+def test_hdf_nonpandas_keys():
+    # https://github.com/dask/dask/issues/5934
+    # TODO: maybe remove this if/when pandas copes with all keys
+
+    tables = pytest.importorskip("tables")
+    import tables
+
+    class Table1(tables.IsDescription):
+        value1 = tables.Float32Col()
+
+    class Table2(tables.IsDescription):
+        value2 = tables.Float32Col()
+
+    class Table3(tables.IsDescription):
+        value3 = tables.Float32Col()
+
+    with tmpfile("h5") as path:
+        with tables.open_file(path, mode="a") as h5file:
+            group = h5file.create_group("/", "group")
+            t = h5file.create_table(group, "table1", Table1, "Table 1")
+            row = t.row
+            row["value1"] = 1
+            row.append()
+            t = h5file.create_table(group, "table2", Table2, "Table 2")
+            row = t.row
+            row["value2"] = 1
+            row.append()
+            t = h5file.create_table(group, "table3", Table3, "Table 3")
+            row = t.row
+            row["value3"] = 1
+            row.append()
+
+        # pandas keys should still work
+        bar = pd.DataFrame(np.random.randn(10, 4))
+        bar.to_hdf(path, "/bar", format="table", mode="a")
+
+        dd.read_hdf(path, "/group/table1")
+        dd.read_hdf(path, "/group/table2")
+        dd.read_hdf(path, "/group/table3")
+        dd.read_hdf(path, "/bar")
