@@ -84,6 +84,25 @@ def test_write_json_basic(orient):
         assert_eq(out, df)
 
 
+def test_to_json_with_get():
+    from dask.multiprocessing import get as mp_get
+
+    flag = [False]
+
+    def my_get(*args, **kwargs):
+        flag[0] = True
+        return mp_get(*args, **kwargs)
+
+    df = pd.DataFrame({"x": ["a", "b", "c", "d"], "y": [1, 2, 3, 4]})
+    ddf = dd.from_pandas(df, npartitions=2)
+
+    with tmpdir() as dn:
+        ddf.to_json(dn, compute_kwargs={"scheduler": my_get})
+        assert flag[0]
+        result = dd.read_json(os.path.join(dn, "*"))
+        assert_eq(result, df, check_index=False)
+
+
 def test_read_json_error():
     with tmpfile("json") as f:
         with pytest.raises(ValueError):
