@@ -1,9 +1,12 @@
+from distutils.version import LooseVersion
+
 import pandas as pd
 import pytest
 import numpy as np
 
+import dask.array as da
 import dask.dataframe as dd
-from dask.dataframe.utils import assert_eq, PANDAS_VERSION
+from dask.dataframe.utils import assert_eq
 
 N = 40
 df = pd.DataFrame(
@@ -41,12 +44,6 @@ def shifted_sum(df, before, after, c=0):
     return df + a + b + c
 
 
-def ts_shifted_sum(df, before, after, c=0):
-    a = df.shift(before.seconds)
-    b = df.shift(-after.seconds)
-    return df + a + b + c
-
-
 @pytest.mark.parametrize("npartitions", [1, 4])
 def test_map_overlap(npartitions):
     ddf = dd.from_pandas(df, npartitions)
@@ -143,6 +140,10 @@ rolling_method_args_check_less_precise = [
 @pytest.mark.parametrize("window", [1, 2, 4, 5])
 @pytest.mark.parametrize("center", [True, False])
 def test_rolling_methods(method, args, window, center, check_less_precise):
+    if dd._compat.PANDAS_GT_110:
+        check_less_precise = {}
+    else:
+        check_less_precise = {"check_less_precise": check_less_precise}
     # DataFrame
     prolling = df.rolling(window, center=center)
     drolling = ddf.rolling(window, center=center)
@@ -154,7 +155,7 @@ def test_rolling_methods(method, args, window, center, check_less_precise):
     assert_eq(
         getattr(prolling, method)(*args, **kwargs),
         getattr(drolling, method)(*args, **kwargs),
-        check_less_precise=check_less_precise,
+        **check_less_precise,
     )
 
     # Series
@@ -163,19 +164,10 @@ def test_rolling_methods(method, args, window, center, check_less_precise):
     assert_eq(
         getattr(prolling, method)(*args, **kwargs),
         getattr(drolling, method)(*args, **kwargs),
-        check_less_precise=check_less_precise,
+        **check_less_precise,
     )
 
 
-if PANDAS_VERSION <= "0.25.0":
-    filter_panel_warning = pytest.mark.filterwarnings(
-        "ignore::DeprecationWarning:pandas[.*]"
-    )
-else:
-    filter_panel_warning = lambda f: f
-
-
-@filter_panel_warning
 @pytest.mark.parametrize("window", [1, 2, 4, 5])
 @pytest.mark.parametrize("center", [True, False])
 def test_rolling_cov(window, center):
@@ -268,6 +260,14 @@ def test_time_rolling_constructor():
 )
 @pytest.mark.parametrize("window", ["1S", "2S", "3S", pd.offsets.Second(5)])
 def test_time_rolling_methods(method, args, window, check_less_precise):
+    if dd._compat.PANDAS_GT_110:
+        if check_less_precise:
+            check_less_precise = {"atol": 0.5e-3, "rtol": 0.5e-3}
+        else:
+            check_less_precise = {}
+    else:
+        check_less_precise = {"check_less_precise": check_less_precise}
+
     # DataFrame
     if method == "apply":
         kwargs = {"raw": False}
@@ -278,7 +278,7 @@ def test_time_rolling_methods(method, args, window, check_less_precise):
     assert_eq(
         getattr(prolling, method)(*args, **kwargs),
         getattr(drolling, method)(*args, **kwargs),
-        check_less_precise=check_less_precise,
+        **check_less_precise,
     )
 
     # Series
@@ -287,11 +287,10 @@ def test_time_rolling_methods(method, args, window, check_less_precise):
     assert_eq(
         getattr(prolling, method)(*args, **kwargs),
         getattr(drolling, method)(*args, **kwargs),
-        check_less_precise=check_less_precise,
+        **check_less_precise,
     )
 
 
-@filter_panel_warning
 @pytest.mark.parametrize("window", ["1S", "2S", "3S", pd.offsets.Second(5)])
 def test_time_rolling_cov(window):
     # DataFrame
@@ -382,8 +381,13 @@ def test_rolling_agg_aggregate():
 
 
 @pytest.mark.skipif(not dd._compat.PANDAS_GT_100, reason="needs pandas>=1.0.0")
+@pytest.mark.xfail(da.numpy_compat._numpy_120, reason="sparse-383")
 def test_rolling_numba_engine():
-    pytest.importorskip("numba")
+    numba = pytest.importorskip("numba")
+    if not dd._compat.PANDAS_GT_104 and LooseVersion(numba.__version__) >= "0.49":
+        # Was fixed in https://github.com/pandas-dev/pandas/pull/33687
+        pytest.xfail("Known incompatibility between pandas and numba")
+
     df = pd.DataFrame({"A": range(5), "B": range(0, 10, 2)})
     ddf = dd.from_pandas(df, npartitions=3)
 
