@@ -21,11 +21,8 @@ from pandas.api.types import (
 
 # include these here for compat
 from ._compat import (  # noqa: F401
-    PANDAS_VERSION,
-    PANDAS_GT_0240,
-    PANDAS_GT_0250,
     PANDAS_GT_100,
-    HAS_INT_NA,
+    PANDAS_GT_110,
     tm,
 )
 
@@ -40,28 +37,26 @@ from ..utils import is_index_like as dask_is_index_like
 
 # register pandas extension types
 from . import _dtypes  # noqa: F401
+from . import methods
 
 
 def is_integer_na_dtype(t):
     dtype = getattr(t, "dtype", t)
-    if HAS_INT_NA:
-        types = (
-            pd.Int8Dtype,
-            pd.Int16Dtype,
-            pd.Int32Dtype,
-            pd.Int64Dtype,
-            pd.UInt8Dtype,
-            pd.UInt16Dtype,
-            pd.UInt32Dtype,
-            pd.UInt64Dtype,
-        )
-    else:
-        types = ()
+    types = (
+        pd.Int8Dtype,
+        pd.Int16Dtype,
+        pd.Int32Dtype,
+        pd.Int64Dtype,
+        pd.UInt8Dtype,
+        pd.UInt16Dtype,
+        pd.UInt32Dtype,
+        pd.UInt64Dtype,
+    )
     return isinstance(dtype, types)
 
 
 def shard_df_on_index(df, divisions):
-    """ Shard a DataFrame by ranges on its index
+    """Shard a DataFrame by ranges on its index
 
     Examples
     --------
@@ -190,7 +185,7 @@ def raise_on_meta_error(funcname=None, udf=False):
             "{2}"
         )
         msg = msg.format(" in `{0}`".format(funcname) if funcname else "", repr(e), tb)
-        raise ValueError(msg)
+        raise ValueError(msg) from e
 
 
 UNKNOWN_CATEGORIES = "__UNKNOWN_CATEGORIES__"
@@ -312,13 +307,14 @@ def make_meta_object(x, index=None):
 
     Examples
     --------
-    >>> make_meta([('a', 'i8'), ('b', 'O')])
+
+    >>> make_meta([('a', 'i8'), ('b', 'O')])    # doctest: +SKIP
     Empty DataFrame
     Columns: [a, b]
     Index: []
-    >>> make_meta(('a', 'f8'))
+    >>> make_meta(('a', 'f8'))                  # doctest: +SKIP
     Series([], Name: a, dtype: float64)
-    >>> make_meta('i8')
+    >>> make_meta('i8')                         # doctest: +SKIP
     1
     """
     if hasattr(x, "_meta"):
@@ -396,6 +392,8 @@ def meta_nonempty_dataframe(x):
         data[i] = dt_s_dict[dt]
     res = pd.DataFrame(data, index=idx, columns=np.arange(len(x.columns)))
     res.columns = x.columns
+    if PANDAS_GT_100:
+        res.attrs = x.attrs
     return res
 
 
@@ -477,11 +475,16 @@ group_split_dispatch = Dispatch("group_split_dispatch")
 
 
 @group_split_dispatch.register((pd.DataFrame, pd.Series, pd.Index))
-def group_split_pandas(df, c, k):
-    indexer, locations = pd._libs.algos.groupsort_indexer(c, k)
+def group_split_pandas(df, c, k, ignore_index=False):
+    indexer, locations = pd._libs.algos.groupsort_indexer(
+        c.astype(np.int64, copy=False), k
+    )
     df2 = df.take(indexer)
     locations = locations.cumsum()
-    parts = [df2.iloc[a:b] for a, b in zip(locations[:-1], locations[1:])]
+    parts = [
+        df2.iloc[a:b].reset_index(drop=True) if ignore_index else df2.iloc[a:b]
+        for a, b in zip(locations[:-1], locations[1:])
+    ]
     return dict(zip(range(k), parts))
 
 
@@ -539,7 +542,10 @@ def _nonempty_series(s, idx=None):
     if idx is None:
         idx = _nonempty_index(s.index)
     dtype = s.dtype
-    if is_datetime64tz_dtype(dtype):
+    if len(s) > 0:
+        # use value from meta if provided
+        data = [s.iloc[0]] * 2
+    elif is_datetime64tz_dtype(dtype):
         entry = pd.Timestamp("1970-01-01", tz=dtype.tz)
         data = [entry, entry]
     elif is_categorical_dtype(dtype):
@@ -548,7 +554,7 @@ def _nonempty_series(s, idx=None):
             cats = s.cat.categories
         else:
             data = _nonempty_index(s.cat.categories)
-            cats = None
+            cats = s.cat.categories[:0]
         data = pd.Categorical(data, categories=cats, ordered=s.cat.ordered)
     elif is_integer_na_dtype(dtype):
         data = pd.array([1, None], dtype=dtype)
@@ -557,26 +563,24 @@ def _nonempty_series(s, idx=None):
         freq = dtype.freq
         data = [pd.Period("2000", freq), pd.Period("2001", freq)]
     elif is_sparse(dtype):
-        # TODO: pandas <0.24
-        # Pandas <= 0.23.4:
-        if PANDAS_GT_0240:
-            entry = _scalar_from_dtype(dtype.subtype)
-        else:
-            entry = _scalar_from_dtype(dtype.subtype)
-        data = pd.SparseArray([entry, entry], dtype=dtype)
-    elif is_interval_dtype(dtype):
         entry = _scalar_from_dtype(dtype.subtype)
-        if PANDAS_GT_0240:
+        if PANDAS_GT_100:
             data = pd.array([entry, entry], dtype=dtype)
         else:
-            data = np.array([entry, entry], dtype=dtype)
+            data = pd.SparseArray([entry, entry], dtype=dtype)
+    elif is_interval_dtype(dtype):
+        entry = _scalar_from_dtype(dtype.subtype)
+        data = pd.array([entry, entry], dtype=dtype)
     elif type(dtype) in make_array_nonempty._lookup:
         data = make_array_nonempty(dtype)
     else:
         entry = _scalar_from_dtype(dtype)
         data = np.array([entry, entry], dtype=dtype)
 
-    return pd.Series(data, name=s.name, index=idx)
+    out = pd.Series(data, name=s.name, index=idx)
+    if PANDAS_GT_100:
+        out.attrs = s.attrs
+    return out
 
 
 def is_dataframe_like(df):
@@ -632,7 +636,9 @@ def check_meta(x, meta, funcname=None, numeric_equal=True):
             "Index, got `%s`" % typename(type(meta))
         )
 
-    if type(x) != type(meta):
+    # Notice, we use .__class__ as opposed to type() in order to support
+    # object proxies see <https://github.com/dask/dask/pull/6981>
+    if x.__class__ != meta.__class__:
         errmsg = "Expected partition of type `%s` but got `%s`" % (
             typename(type(meta)),
             typename(type(x)),
@@ -640,7 +646,7 @@ def check_meta(x, meta, funcname=None, numeric_equal=True):
     elif is_dataframe_like(meta):
         dtypes = pd.concat([x.dtypes, meta.dtypes], axis=1, sort=True)
         bad_dtypes = [
-            (col, a, b)
+            (repr(col), a, b)
             for col, a, b in dtypes.fillna("-").itertuples()
             if not equal_dtypes(a, b)
         ]
@@ -669,19 +675,21 @@ def check_meta(x, meta, funcname=None, numeric_equal=True):
 def check_matching_columns(meta, actual):
     # Need nan_to_num otherwise nan comparison gives False
     if not np.array_equal(np.nan_to_num(meta.columns), np.nan_to_num(actual.columns)):
-        extra = actual.columns.difference(meta.columns).tolist()
-        missing = meta.columns.difference(actual.columns).tolist()
+        extra = methods.tolist(actual.columns.difference(meta.columns))
+        missing = methods.tolist(meta.columns.difference(actual.columns))
+        if extra or missing:
+            extra_info = f"  Extra:   {extra}\n  Missing: {missing}"
+        else:
+            extra_info = "Order of columns does not match"
         raise ValueError(
             "The columns in the computed data do not match"
             " the columns in the provided metadata\n"
-            "  Extra:   %s\n"
-            "  Missing: %s" % (extra, missing)
+            f"{extra_info}"
         )
 
 
 def index_summary(idx, name=None):
-    """Summarized representation of an Index.
-    """
+    """Summarized representation of an Index."""
     n = len(idx)
     if name is None:
         name = idx.__class__.__name__
@@ -703,7 +711,10 @@ def index_summary(idx, name=None):
 def _check_dask(dsk, check_names=True, check_dtypes=True, result=None):
     import dask.dataframe as dd
 
-    if hasattr(dsk, "dask"):
+    if hasattr(dsk, "__dask_graph__"):
+        graph = dsk.__dask_graph__()
+        if hasattr(graph, "validate"):
+            graph.validate()
         if result is None:
             result = dsk.compute(scheduler="sync")
         if isinstance(dsk, dd.Index):
@@ -766,7 +777,7 @@ def _maybe_sort(a):
                 a.index.names = [
                     "-overlapped-index-name-%d" % i for i in range(len(a.index.names))
                 ]
-            a = a.sort_values(by=a.columns.tolist())
+            a = a.sort_values(by=methods.tolist(a.columns))
         else:
             a = a.sort_values()
     except (TypeError, IndexError, ValueError):
@@ -781,7 +792,7 @@ def assert_eq(
     check_dtypes=True,
     check_divisions=True,
     check_index=True,
-    **kwargs
+    **kwargs,
 ):
     if check_divisions:
         assert_divisions(a)
@@ -918,7 +929,7 @@ def assert_max_deps(x, n, eq=True):
 
 
 def valid_divisions(divisions):
-    """ Are the provided divisions valid?
+    """Are the provided divisions valid?
 
     Examples
     --------
@@ -955,8 +966,7 @@ def valid_divisions(divisions):
 
 
 def drop_by_shallow_copy(df, columns, errors="raise"):
-    """ Use shallow copy to drop columns in place
-    """
+    """Use shallow copy to drop columns in place"""
     df2 = df.copy(deep=False)
     if not pd.api.types.is_list_like(columns):
         columns = [columns]
