@@ -1,26 +1,23 @@
-from __future__ import absolute_import, division, print_function
-
 import operator
 import types
 import uuid
+import warnings
+from collections.abc import Iterator
 
-try:
-    from cytoolz import curry, pluck
-except ImportError:
-    from toolz import curry, pluck
+from tlz import curry, concat, unique, merge
 
 from . import config, threaded
 from .base import is_dask_collection, dont_optimize, DaskMethodsMixin
 from .base import tokenize as _tokenize
-from .compatibility import apply, Iterator, is_dataclass, dataclass_fields
+from .compatibility import is_dataclass, dataclass_fields
 
 from .core import quote
 from .context import globalmethod
 from .optimization import cull
-from .utils import funcname, methodcaller, OperatorMethodMixin, ensure_dict
-from . import sharedict
+from .utils import funcname, methodcaller, OperatorMethodMixin, ensure_dict, apply
+from .highlevelgraph import HighLevelGraph
 
-__all__ = ['Delayed', 'delayed']
+__all__ = ["Delayed", "delayed"]
 
 
 def unzip(ls, nout):
@@ -31,6 +28,91 @@ def unzip(ls, nout):
     return out
 
 
+def finalize(collection):
+    assert is_dask_collection(collection)
+
+    name = "finalize-" + tokenize(collection)
+    keys = collection.__dask_keys__()
+    finalize, args = collection.__dask_postcompute__()
+    layer = {name: (finalize, keys) + args}
+    graph = HighLevelGraph.from_collections(name, layer, dependencies=[collection])
+    return Delayed(name, graph)
+
+
+def unpack_collections(expr):
+    """Normalize a python object and merge all sub-graphs.
+
+    - Replace ``Delayed`` with their keys
+    - Convert literals to things the schedulers can handle
+    - Extract dask graphs from all enclosed values
+
+    Parameters
+    ----------
+    expr : object
+        The object to be normalized. This function knows how to handle
+        dask collections, as well as most builtin python types.
+
+    Returns
+    -------
+    task : normalized task to be run
+    collections : a tuple of collections
+
+    Examples
+    --------
+    >>> import dask
+    >>> a = delayed(1, 'a')
+    >>> b = delayed(2, 'b')
+    >>> task, collections = unpack_collections([a, b, 3])
+    >>> task  # doctest: +SKIP
+    ['a', 'b', 3]
+    >>> collections  # doctest: +SKIP
+    (a, b)
+
+    >>> task, collections = unpack_collections({a: 1, b: 2})
+    >>> task  # doctest: +SKIP
+    (dict, [['a', 1], ['b', 2]])
+    >>> collections  # doctest: +SKIP
+    {a, b}
+    """
+    if isinstance(expr, Delayed):
+        return expr._key, (expr,)
+
+    if is_dask_collection(expr):
+        finalized = finalize(expr)
+        return finalized._key, (finalized,)
+
+    if isinstance(expr, Iterator):
+        expr = tuple(expr)
+
+    typ = type(expr)
+
+    if typ in (list, tuple, set):
+        args, collections = unzip((unpack_collections(e) for e in expr), 2)
+        args = list(args)
+        collections = tuple(unique(concat(collections), key=id))
+        # Ensure output type matches input type
+        if typ is not list:
+            args = (typ, args)
+        return args, collections
+
+    if typ is dict:
+        args, collections = unpack_collections([[k, v] for k, v in expr.items()])
+        return (dict, args), collections
+
+    if typ is slice:
+        args, collections = unpack_collections([expr.start, expr.stop, expr.step])
+        return (slice,) + tuple(args), collections
+
+    if is_dataclass(expr):
+        args, collections = unpack_collections(
+            [[f.name, getattr(expr, f.name)] for f in dataclass_fields(expr)]
+        )
+
+        return (apply, typ, (), (dict, args)), collections
+
+    return expr, ()
+
+
 def to_task_dask(expr):
     """Normalize a python object and merge all sub-graphs.
 
@@ -51,27 +133,34 @@ def to_task_dask(expr):
 
     Examples
     --------
+    >>> import dask
     >>> a = delayed(1, 'a')
     >>> b = delayed(2, 'b')
-    >>> task, dask = to_task_dask([a, b, 3])
+    >>> task, dask = to_task_dask([a, b, 3])  # doctest: +SKIP
     >>> task  # doctest: +SKIP
     ['a', 'b', 3]
     >>> dict(dask)  # doctest: +SKIP
     {'a': 1, 'b': 2}
 
-    >>> task, dasks = to_task_dask({a: 1, b: 2})
+    >>> task, dasks = to_task_dask({a: 1, b: 2})  # doctest: +SKIP
     >>> task  # doctest: +SKIP
     (dict, [['a', 1], ['b', 2]])
     >>> dict(dask)  # doctest: +SKIP
     {'a': 1, 'b': 2}
     """
+    warnings.warn(
+        "The dask.delayed.to_dask_dask function has been "
+        "Deprecated in favor of unpack_collections",
+        stacklevel=2,
+    )
+
     if isinstance(expr, Delayed):
         return expr.key, expr.dask
 
     if is_dask_collection(expr):
-        name = 'finalize-' + tokenize(expr, pure=True)
+        name = "finalize-" + tokenize(expr, pure=True)
         keys = expr.__dask_keys__()
-        opt = getattr(expr, '__dask_optimize__', dont_optimize)
+        opt = getattr(expr, "__dask_optimize__", dont_optimize)
         finalize, args = expr.__dask_postcompute__()
         dsk = {name: (finalize, keys) + args}
         dsk.update(opt(expr.__dask_graph__(), keys))
@@ -84,7 +173,7 @@ def to_task_dask(expr):
     if typ in (list, tuple, set):
         args, dasks = unzip((to_task_dask(e) for e in expr), 2)
         args = list(args)
-        dsk = sharedict.merge(*dasks, dependencies={})  # TODO
+        dsk = merge(dasks)
         # Ensure output type matches input type
         return (args, dsk) if typ is list else ((typ, args), dsk)
 
@@ -93,8 +182,9 @@ def to_task_dask(expr):
         return (dict, args), dsk
 
     if is_dataclass(expr):
-        args, dsk = to_task_dask([[f.name, getattr(expr, f.name)] for f in
-                                  dataclass_fields(expr)])
+        args, dsk = to_task_dask(
+            [[f.name, getattr(expr, f.name)] for f in dataclass_fields(expr)]
+        )
 
         return (apply, typ, (), (dict, args)), dsk
 
@@ -117,9 +207,9 @@ def tokenize(*args, **kwargs):
         fails, then a unique identifier is used. If False (default), then a
         unique identifier is always used.
     """
-    pure = kwargs.pop('pure', None)
+    pure = kwargs.pop("pure", None)
     if pure is None:
-        pure = config.get('delayed_pure', False)
+        pure = config.get("delayed_pure", False)
 
     if pure:
         return _tokenize(*args, **kwargs)
@@ -143,6 +233,15 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
         to hashing content. Note that this only affects the name of the object
         wrapped by this call to delayed, and *not* the output of delayed
         function calls - for that use ``dask_key_name=`` as described below.
+
+        .. note::
+
+           Because this ``name`` is used as the key in task graphs, you should
+           ensure that it uniquely identifies ``obj``. If you'd like to provide
+           a descriptive name that is still unique, combine the descriptive name
+           with :func:`dask.base.tokenize` of the ``array_like``. See
+           :ref:`graphs` for more.
+
     pure : bool, optional
         Indicates whether calling the resulting ``Delayed`` object is a pure
         operation. If True, arguments to the call are hashed to produce
@@ -153,7 +252,7 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
         object. If provided, the ``Delayed`` output of the call can be iterated
         into ``nout`` objects, allowing for unpacking of results. By default
         iteration over ``Delayed`` objects will error. Note, that ``nout=1``
-        expects ``obj``, to return a tuple of length 1, and consequently for
+        expects ``obj`` to return a tuple of length 1, and consequently for
         ``nout=0``, ``obj`` should return an empty tuple.
     traverse : bool, optional
         By default dask traverses builtin python collections looking for dask
@@ -165,6 +264,7 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
     --------
     Apply to functions to delay execution:
 
+    >>> from dask import delayed
     >>> def inc(x):
     ...     return x + 1
 
@@ -172,7 +272,7 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
     11
 
     >>> x = delayed(inc, pure=True)(10)
-    >>> type(x) == Delayed
+    >>> type(x) == Delayed              # doctest: +SKIP
     True
     >>> x.compute()
     11
@@ -212,10 +312,10 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
     it contextually using the ``delayed_pure`` setting. Note that this
     influences the *call* and not the *creation* of the callable:
 
-    >>> import dask
     >>> @delayed
     ... def mul(a, b):
     ...     return a * b
+    >>> import dask
     >>> with dask.config.set(delayed_pure=True):
     ...     print(mul(1, 2).key == mul(1, 2).key)
     True
@@ -227,7 +327,7 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
     hashing the arguments by default. To explicitly set the name, you can use
     the ``dask_key_name`` keyword when calling the function:
 
-    >>> add(1, 2)    # doctest: +SKIP
+    >>> add(1, 2)   # doctest: +SKIP
     Delayed('add-3dce7c56edd1ac2614add714086e950f')
     >>> add(1, 2, dask_key_name='three')
     Delayed('three')
@@ -236,42 +336,49 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
     result. If you set the names explicitly you should make sure your key names
     are different for different results.
 
-    >>> add(1, 2, dask_key_name='three')  # doctest: +SKIP
-    >>> add(2, 1, dask_key_name='three')  # doctest: +SKIP
-    >>> add(2, 2, dask_key_name='four')   # doctest: +SKIP
+    >>> add(1, 2, dask_key_name='three')
+    Delayed('three')
+    >>> add(2, 1, dask_key_name='three')
+    Delayed('three')
+    >>> add(2, 2, dask_key_name='four')
+    Delayed('four')
 
     ``delayed`` can also be applied to objects to make operations on them lazy:
 
     >>> a = delayed([1, 2, 3])
-    >>> isinstance(a, Delayed)
+    >>> isinstance(a, Delayed)      # doctest: +SKIP
     True
     >>> a.compute()
     [1, 2, 3]
 
     The key name of a delayed object is hashed by default if ``pure=True`` or
     is generated randomly if ``pure=False`` (default).  To explicitly set the
-    name, you can use the ``name`` keyword:
+    name, you can use the ``name`` keyword. To ensure that the key is unique
+    you should include the tokenized value as well, or otherwise ensure that
+    it's unique:
 
-    >>> a = delayed([1, 2, 3], name='mylist')
-    >>> a
-    Delayed('mylist')
+    >>> from dask.base import tokenize
+    >>> data = [1, 2, 3]
+    >>> a = delayed(data, name='mylist-' + tokenize(data))
+    >>> a  # doctest: +SKIP
+    Delayed('mylist-55af65871cb378a4fa6de1660c3e8fb7')
 
     Delayed results act as a proxy to the underlying object. Many operators
     are supported:
 
-    >>> (a + [1, 2]).compute()
+    >>> (a + [1, 2]).compute()      # doctest: +SKIP
     [1, 2, 3, 1, 2]
-    >>> a[1].compute()
+    >>> a[1].compute()              # doctest: +SKIP
     2
 
     Method and attribute access also works:
 
-    >>> a.count(2).compute()
+    >>> a.count(2).compute()        # doctest: +SKIP
     1
 
     Note that if a method doesn't exist, no error will be thrown until runtime:
 
-    >>> res = a.not_a_real_method()
+    >>> res = a.not_a_real_method() # doctest: +SKIP
     >>> res.compute()  # doctest: +SKIP
     AttributeError("'list' object has no attribute 'not_a_real_method'")
 
@@ -300,7 +407,7 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
 
     In contrast, methods are assumed to be impure by default, meaning that
     subsequent calls may return different results. To assume purity, set
-    `pure=True`. This allows sharing of any intermediate values.
+    ``pure=True``. This allows sharing of any intermediate values.
 
     >>> a.count(2, pure=True).key == a.count(2, pure=True).key
     True
@@ -310,6 +417,7 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
 
     >>> a.count(2, dask_key_name="count_2")
     Delayed('count_2')
+    >>> import dask
     >>> with dask.config.set(delayed_pure=True):
     ...     print(a.count(2).key == a.count(2).key)
     True
@@ -318,34 +426,38 @@ def delayed(obj, name=None, pure=None, nout=None, traverse=True):
         return obj
 
     if is_dask_collection(obj) or traverse:
-        task, dsk = to_task_dask(obj)
+        task, collections = unpack_collections(obj)
     else:
         task = quote(obj)
-        dsk = {}
+        collections = set()
 
     if task is obj:
         if not (nout is None or (type(nout) is int and nout >= 0)):
-            raise ValueError("nout must be None or a non-negative integer,"
-                             " got %s" % nout)
+            raise ValueError(
+                "nout must be None or a non-negative integer, got %s" % nout
+            )
         if not name:
             try:
                 prefix = obj.__name__
             except AttributeError:
                 prefix = type(obj).__name__
             token = tokenize(obj, nout, pure=pure)
-            name = '%s-%s' % (prefix, token)
+            name = "%s-%s" % (prefix, token)
         return DelayedLeaf(obj, name, pure=pure, nout=nout)
     else:
         if not name:
-            name = '%s-%s' % (type(obj).__name__, tokenize(task, pure=pure))
-        dsk = sharedict.merge(dsk, (name, {name: task}), dependencies={})
-        return Delayed(name, dsk)
+            name = "%s-%s" % (type(obj).__name__, tokenize(task, pure=pure))
+        layer = {name: task}
+        graph = HighLevelGraph.from_collections(name, layer, dependencies=collections)
+        return Delayed(name, graph)
 
 
 def right(method):
     """Wrapper to create 'right' version of operator given left version"""
+
     def _inner(self, other):
         return method(other, self)
+
     return _inner
 
 
@@ -364,32 +476,44 @@ class Delayed(DaskMethodsMixin, OperatorMethodMixin):
 
     Equivalent to the output from a single key in a dask graph.
     """
-    __slots__ = ('_key', 'dask', '_length')
+
+    __slots__ = ("_key", "dask", "_length", "_dask_layers")
 
     def __init__(self, key, dsk, length=None):
         self._key = key
-        if type(dsk) is list:  # compatibility with older versions
-            dsk = sharedict.merge(*dsk, dependencies={})
         self.dask = dsk
         self._length = length
 
+        # `__dask_layers__` of an one-layered HLG is the layer name
+        # TODO: what is the name of a multi-layered HLG?
+        if isinstance(dsk, HighLevelGraph) and len(dsk.layers) == 1:
+            self._dask_layers = set(iter(dsk.layers))
+        else:
+            self._dask_layers = (self.key,)
+
     def __dask_graph__(self):
         return self.dask
 
     def __dask_keys__(self):
         return [self.key]
 
+    def __dask_layers__(self):
+        if hasattr(self, "_dask_layers"):
+            return self._dask_layers
+        else:
+            return (self.key,)
+
     def __dask_tokenize__(self):
         return self.key
 
     __dask_scheduler__ = staticmethod(threaded.get)
-    __dask_optimize__ = globalmethod(optimize, key='delayed_optimize')
+    __dask_optimize__ = globalmethod(optimize, key="delayed_optimize")
 
     def __dask_postcompute__(self):
         return single_key, ()
 
     def __dask_postpersist__(self):
-        return rebuild, (self._key, getattr(self, '_length', None))
+        return rebuild, (self._key, getattr(self, "_length", None))
 
     def __getstate__(self):
         return tuple(getattr(self, i) for i in self.__slots__)
@@ -412,8 +536,16 @@ class Delayed(DaskMethodsMixin, OperatorMethodMixin):
         return dir(type(self))
 
     def __getattr__(self, attr):
-        if attr.startswith('_'):
+        if attr.startswith("_"):
             raise AttributeError("Attribute {0} not found".format(attr))
+
+        if attr == "visualise":
+            # added to warn users in case of spelling error
+            # for more details: https://github.com/dask/dask/issues/5721
+            warnings.warn(
+                "dask.delayed objects have no `visualise` method. Perhaps you meant `visualize`?"
+            )
+
         return DelayedAttr(self, attr)
 
     def __setattr__(self, attr, val):
@@ -426,21 +558,19 @@ class Delayed(DaskMethodsMixin, OperatorMethodMixin):
         raise TypeError("Delayed objects are immutable")
 
     def __iter__(self):
-        if getattr(self, '_length', None) is None:
-            raise TypeError("Delayed objects of unspecified length are "
-                            "not iterable")
+        if getattr(self, "_length", None) is None:
+            raise TypeError("Delayed objects of unspecified length are not iterable")
         for i in range(self._length):
             yield self[i]
 
     def __len__(self):
-        if getattr(self, '_length', None) is None:
-            raise TypeError("Delayed objects of unspecified length have "
-                            "no len()")
+        if getattr(self, "_length", None) is None:
+            raise TypeError("Delayed objects of unspecified length have no len()")
         return self._length
 
     def __call__(self, *args, **kwargs):
-        pure = kwargs.pop('pure', None)
-        name = kwargs.pop('dask_key_name', None)
+        pure = kwargs.pop("pure", None)
+        name = kwargs.pop("dask_key_name", None)
         func = delayed(apply, pure=pure)
         if name is not None:
             return func(self, args, kwargs, dask_key_name=name)
@@ -465,41 +595,36 @@ class Delayed(DaskMethodsMixin, OperatorMethodMixin):
 
 
 def call_function(func, func_token, args, kwargs, pure=None, nout=None):
-    dask_key_name = kwargs.pop('dask_key_name', None)
-    pure = kwargs.pop('pure', pure)
+    dask_key_name = kwargs.pop("dask_key_name", None)
+    pure = kwargs.pop("pure", pure)
 
     if dask_key_name is None:
-        name = '%s-%s' % (funcname(func),
-                          tokenize(func_token, *args, pure=pure, **kwargs))
+        name = "%s-%s" % (
+            funcname(func),
+            tokenize(func_token, *args, pure=pure, **kwargs),
+        )
     else:
         name = dask_key_name
 
-    dsk = sharedict.ShareDict()
-    args_dasks = list(map(to_task_dask, args))
-    for arg, d in args_dasks:
-        if isinstance(d, sharedict.ShareDict):
-            dsk.update_with_key(d)
-        elif isinstance(arg, (str, tuple)):
-            dsk.update_with_key(d, key=arg)
-        else:
-            dsk.update(d)
-
-    args = tuple(pluck(0, args_dasks))
+    args2, collections = unzip(map(unpack_collections, args), 2)
+    collections = list(concat(collections))
 
     if kwargs:
-        dask_kwargs, dsk2 = to_task_dask(kwargs)
-        dsk.update(dsk2)
-        task = (apply, func, list(args), dask_kwargs)
+        dask_kwargs, collections2 = unpack_collections(kwargs)
+        collections.extend(collections2)
+        task = (apply, func, list(args2), dask_kwargs)
     else:
-        task = (func,) + args
+        task = (func,) + args2
 
-    dsk.update_with_key({name: task}, key=name)
+    graph = HighLevelGraph.from_collections(
+        name, {name: task}, dependencies=collections
+    )
     nout = nout if nout is not None else None
-    return Delayed(name, dsk, length=nout)
+    return Delayed(name, graph, length=nout)
 
 
 class DelayedLeaf(Delayed):
-    __slots__ = ('_obj', '_key', '_pure', '_nout')
+    __slots__ = ("_obj", "_key", "_pure", "_nout")
 
     def __init__(self, obj, key, pure=None, nout=None):
         self._obj = obj
@@ -509,38 +634,72 @@ class DelayedLeaf(Delayed):
 
     @property
     def dask(self):
-        return sharedict.ShareDict({self._key: {self._key: self._obj}},
-                                   dependencies={self._key: set()})
+        return HighLevelGraph.from_collections(
+            self._key, {self._key: self._obj}, dependencies=()
+        )
 
     def __call__(self, *args, **kwargs):
-        return call_function(self._obj, self._key, args, kwargs,
-                             pure=self._pure, nout=self._nout)
+        return call_function(
+            self._obj, self._key, args, kwargs, pure=self._pure, nout=self._nout
+        )
 
 
 class DelayedAttr(Delayed):
-    __slots__ = ('_obj', '_attr', '_key')
+    __slots__ = ("_obj", "_attr", "_key")
 
     def __init__(self, obj, attr):
         self._obj = obj
         self._attr = attr
-        self._key = 'getattr-%s' % tokenize(obj, attr, pure=True)
+        self._key = "getattr-%s" % tokenize(obj, attr, pure=True)
+
+    def __getattr__(self, attr):
+        # Calling np.dtype(dask.delayed(...)) used to result in a segfault, as
+        # numpy recursively tries to get `dtype` from the object. This is
+        # likely a bug in numpy. For now, we can do a dumb for if
+        # `x.dtype().dtype()` is called (which shouldn't ever show up in real
+        # code). See https://github.com/dask/dask/pull/4374#issuecomment-454381465
+        if attr == "dtype" and self._attr == "dtype":
+            raise AttributeError("Attribute %s not found" % attr)
+        return super().__getattr__(attr)
 
     @property
     def dask(self):
-        dsk = {self._key: (getattr, self._obj._key, self._attr)}
-        return sharedict.merge(self._obj.dask, (self._key, dsk),
-                               dependencies={})
+        layer = {self._key: (getattr, self._obj._key, self._attr)}
+        return HighLevelGraph.from_collections(
+            self._key, layer, dependencies=[self._obj]
+        )
 
     def __call__(self, *args, **kwargs):
-        return call_function(methodcaller(self._attr), self._attr, (self._obj,) + args, kwargs)
+        return call_function(
+            methodcaller(self._attr), self._attr, (self._obj,) + args, kwargs
+        )
 
 
-for op in [operator.abs, operator.neg, operator.pos, operator.invert,
-           operator.add, operator.sub, operator.mul, operator.floordiv,
-           operator.truediv, operator.mod, operator.pow, operator.and_,
-           operator.or_, operator.xor, operator.lshift, operator.rshift,
-           operator.eq, operator.ge, operator.gt, operator.ne, operator.le,
-           operator.lt, operator.getitem]:
+for op in [
+    operator.abs,
+    operator.neg,
+    operator.pos,
+    operator.invert,
+    operator.add,
+    operator.sub,
+    operator.mul,
+    operator.floordiv,
+    operator.truediv,
+    operator.mod,
+    operator.pow,
+    operator.and_,
+    operator.or_,
+    operator.xor,
+    operator.lshift,
+    operator.rshift,
+    operator.eq,
+    operator.ge,
+    operator.gt,
+    operator.ne,
+    operator.le,
+    operator.lt,
+    operator.getitem,
+]:
     Delayed._bind_operator(op)
 
 
