@@ -1,15 +1,13 @@
+from distutils.version import LooseVersion
 import sys
 import multiprocessing
 from operator import add
 import pickle
-import random
-
-import numpy as np
 
 import pytest
 import dask
 from dask import compute, delayed
-from dask.multiprocessing import get, _dumps, get_context, remote_exception
+from dask.multiprocessing import get, _dumps, _loads, get_context, remote_exception
 from dask.utils_test import inc
 
 
@@ -27,6 +25,7 @@ not_cloudpickle = pytest.mark.skipif(has_cloudpickle, reason="cloudpickle is ins
 
 
 def unrelated_function_global(a):
+    np = pytest.importorskip("numpy")
     return np.array([a])
 
 
@@ -44,8 +43,8 @@ def test_pickle_globals():
 
 @requires_cloudpickle
 def test_pickle_locals():
-    """Unrelated locals should not be included in serialized bytes
-    """
+    """Unrelated locals should not be included in serialized bytes"""
+    np = pytest.importorskip("numpy")
 
     def unrelated_function_local(a):
         return np.array([a])
@@ -59,6 +58,42 @@ def test_pickle_locals():
     assert b"unrelated_function_local" not in b
 
 
+@not_cloudpickle
+def test_pickle_kwargs():
+    """Test that out-of-band pickling works
+
+    Note cloudpickle does not support this argument:
+
+    https://github.com/cloudpipe/cloudpickle/issues/213
+    """
+    b = _dumps(my_small_function_global, fix_imports=True)
+    assert b"my_small_function_global" in b
+    assert b"unrelated_function_global" not in b
+    assert b"numpy" not in b
+    my_small_function_global_2 = _loads(b, fix_imports=True)
+    assert my_small_function_global_2(2, 3) == 5
+
+
+@pytest.mark.skipif(pickle.HIGHEST_PROTOCOL < 5, reason="requires pickle protocol 5")
+def test_out_of_band_pickling():
+    """Test that out-of-band pickling works"""
+    np = pytest.importorskip("numpy")
+    if has_cloudpickle:
+        if cloudpickle.__version__ < LooseVersion("1.3.0"):
+            pytest.skip("when using cloudpickle, it must be version 1.3.0+")
+
+    a = np.arange(5)
+
+    l = []
+    b = _dumps(a, buffer_callback=l.append)
+    assert len(l) == 1
+    assert isinstance(l[0], pickle.PickleBuffer)
+    assert memoryview(l[0]) == memoryview(a)
+
+    a2 = _loads(b, buffers=l)
+    assert np.all(a == a2)
+
+
 def bad():
     raise ValueError("12345")
 
@@ -166,8 +201,14 @@ def test_optimize_graph_false():
 
 
 @requires_cloudpickle
-@pytest.mark.parametrize("random", [np.random, random])
+@pytest.mark.parametrize("random", ["numpy", "random"])
 def test_random_seeds(random):
+    if random == "numpy":
+        np = pytest.importorskip("numpy")
+        random = np.random
+    else:
+        import random
+
     @delayed(pure=False)
     def f():
         return tuple(random.randint(0, 10000) for i in range(5))
@@ -192,17 +233,26 @@ def check_for_pytest():
     sys.platform == "win32", reason="Windows doesn't support different contexts"
 )
 def test_custom_context_used_python3_posix():
-    """ The 'multiprocessing.context' config is used to create the pool.
+    """The 'multiprocessing.context' config is used to create the pool.
 
-    We assume default is 'fork', and therefore test for 'spawn'.  If default
-    context is changed this test will need to be modified to be different than
-    that.
+    We assume default is 'spawn', and therefore test for 'fork'.
     """
+    pytest.importorskip("cloudpickle")
+    # We check for 'fork' by ensuring subprocess doesn't have modules only
+    # parent process should have:
+
+    def check_for_pytest():
+        import sys
+
+        return "FAKE_MODULE_FOR_TEST" in sys.modules
+
+    import sys
+
     sys.modules["FAKE_MODULE_FOR_TEST"] = 1
     try:
-        with dask.config.set({"multiprocessing.context": "spawn"}):
+        with dask.config.set({"multiprocessing.context": "fork"}):
             result = get({"x": (check_for_pytest,)}, "x")
-        assert not result
+        assert result
     finally:
         del sys.modules["FAKE_MODULE_FOR_TEST"]
 
@@ -211,22 +261,22 @@ def test_custom_context_used_python3_posix():
     sys.platform == "win32", reason="Windows doesn't support different contexts"
 )
 def test_get_context_using_python3_posix():
-    """ get_context() respects configuration.
+    """get_context() respects configuration.
 
     If default context is changed this test will need to change too.
     """
-    assert get_context() is multiprocessing.get_context(None)
+    assert get_context() is multiprocessing.get_context("spawn")
     with dask.config.set({"multiprocessing.context": "forkserver"}):
         assert get_context() is multiprocessing.get_context("forkserver")
-    with dask.config.set({"multiprocessing.context": "spawn"}):
-        assert get_context() is multiprocessing.get_context("spawn")
+    with dask.config.set({"multiprocessing.context": "fork"}):
+        assert get_context() is multiprocessing.get_context("fork")
 
 
 @pytest.mark.skipif(sys.platform != "win32", reason="POSIX supports different contexts")
 def test_custom_context_ignored_elsewhere():
-    """ On Windows, setting 'multiprocessing.context' doesn't explode.
+    """On Windows, setting 'multiprocessing.context' doesn't explode.
 
-    Presumption is it's not used since unsupported, but mostly we care about
+    Presumption is it's not used since it's unsupported, but mostly we care about
     not breaking anything.
     """
     assert get({"x": (inc, 1)}, "x") == 2
