@@ -3,9 +3,9 @@ import functools
 import operator
 import pickle
 
-import numpy as np
 import pytest
 
+from dask import get
 from dask.utils import (
     getargspec,
     takes_multiple_arguments,
@@ -28,9 +28,13 @@ from dask.utils import (
     parse_timedelta,
     parse_bytes,
     is_arraylike,
+    iter_chunks,
+    stringify,
+    stringify_collection_keys,
 )
 from dask.utils_test import inc
 from dask.highlevelgraph import HighLevelGraph
+from dask.optimization import SubgraphCallable
 
 
 def test_getargspec():
@@ -152,6 +156,7 @@ def test_dispatch_lazy():
 
 
 def test_random_state_data():
+    np = pytest.importorskip("numpy")
     seed = 37
     state = np.random.RandomState(seed)
     n = 10000
@@ -358,7 +363,7 @@ def test_funcname_long():
 
 
 def test_funcname_toolz():
-    toolz = pytest.importorskip("toolz")
+    toolz = pytest.importorskip("tlz")
 
     @toolz.curry
     def foo(a, b, c):
@@ -454,7 +459,7 @@ def test_has_keyword():
 def test_derived_from():
     class Foo:
         def f(a, b):
-            """ A super docstring
+            """A super docstring
 
             An explanation
 
@@ -525,6 +530,7 @@ def test_parse_bytes():
     assert parse_bytes("1e6 kB") == 1000000000
     assert parse_bytes("MB") == 1000000
     assert parse_bytes(123) == 123
+    assert parse_bytes(".5GB") == 500000000
 
 
 def test_parse_timedelta():
@@ -556,6 +562,8 @@ def test_parse_timedelta():
 
 
 def test_is_arraylike():
+    np = pytest.importorskip("numpy")
+
     assert is_arraylike(0) is False
     assert is_arraylike(()) is False
     assert is_arraylike(0) is False
@@ -565,3 +573,67 @@ def test_is_arraylike():
     assert is_arraylike(np.empty(())) is True
     assert is_arraylike(np.empty((0,))) is True
     assert is_arraylike(np.empty((0, 0))) is True
+
+
+def test_iter_chunks():
+    sizes = [14, 8, 5, 9, 7, 9, 1, 19, 8, 19]
+    assert list(iter_chunks(sizes, 19)) == [
+        [14],
+        [8, 5],
+        [9, 7],
+        [9, 1],
+        [19],
+        [8],
+        [19],
+    ]
+    assert list(iter_chunks(sizes, 28)) == [[14, 8, 5], [9, 7, 9, 1], [19, 8], [19]]
+    assert list(iter_chunks(sizes, 67)) == [[14, 8, 5, 9, 7, 9, 1], [19, 8, 19]]
+
+
+def test_stringify():
+    obj = "Hello"
+    assert stringify(obj) is obj
+    obj = b"Hello"
+    assert stringify(obj) is obj
+    dsk = {"x": 1}
+
+    assert stringify(dsk) == str(dsk)
+    assert stringify(dsk, exclusive=()) == dsk
+
+    dsk = {("x", 1): (inc, 1)}
+    assert stringify(dsk) == str({("x", 1): (inc, 1)})
+    assert stringify(dsk, exclusive=()) == {("x", 1): (inc, 1)}
+
+    dsk = {("x", 1): (inc, 1), ("x", 2): (inc, ("x", 1))}
+    assert stringify(dsk, exclusive=dsk) == {
+        ("x", 1): (inc, 1),
+        ("x", 2): (inc, str(("x", 1))),
+    }
+
+    dsks = [
+        {"x": 1},
+        {("x", 1): (inc, 1), ("x", 2): (inc, ("x", 1))},
+        {("x", 1): (sum, [1, 2, 3]), ("x", 2): (sum, [("x", 1), ("x", 1)])},
+    ]
+    for dsk in dsks:
+        sdsk = {stringify(k): stringify(v, exclusive=dsk) for k, v in dsk.items()}
+        keys = list(dsk)
+        skeys = [str(k) for k in keys]
+        assert all(isinstance(k, str) for k in sdsk)
+        assert get(dsk, keys) == get(sdsk, skeys)
+
+    dsk = {("y", 1): (SubgraphCallable({"x": ("y", 1)}, "x", (("y", 1),)), (("z", 1),))}
+    dsk = stringify(dsk, exclusive=set(dsk) | {("z", 1)})
+    assert dsk[("y", 1)][0].dsk["x"] == "('y', 1)"
+    assert dsk[("y", 1)][1][0] == "('z', 1)"
+
+
+def test_stringify_collection_keys():
+    obj = "Hello"
+    assert stringify_collection_keys(obj) is obj
+
+    obj = [("a", 0), (b"a", 0), (1, 1)]
+    res = stringify_collection_keys(obj)
+    assert res[0] == str(obj[0])
+    assert res[1] == str(obj[1])
+    assert res[2] == obj[2]
